#
# Copyright (C) 2009-2022 Red Hat Inc.
#
# Authors:
#  Luiz Capitulino <lcapitulino@redhat.com>
#  John Snow <jsnow@redhat.com>
#
# This work is licensed under the terms of the GNU LGPL, version 2 or
# later. See the COPYING file in the top-level directory.
#

"""
Low-level QEMU shell on top of QMP.

usage: qmp-shell [-h] [-H] [-N] [-v] [-p] qmp_server

positional arguments:
  qmp_server            < UNIX socket path | TCP address:port >

optional arguments:
  -h, --help            show this help message and exit
  -H, --hmp             Use HMP interface
  -N, --skip-negotiation
                        Skip negotiate (for qemu-ga)
  -v, --verbose         Verbose (echo commands sent and received)
  -p, --pretty          Pretty-print JSON


Start QEMU with:

# qemu [...] -qmp unix:./qmp-sock,server

Run the shell:

$ qmp-shell ./qmp-sock

Commands have the following format:

   < command-name > [ arg-name1=arg1 ] ... [ arg-nameN=argN ]

For example:

(QEMU) device_add driver=e1000 id=net1
{'return': {}}
(QEMU)

key=value pairs also support Python or JSON object literal subset notations,
without spaces. Dictionaries/objects {} are supported as are arrays [].

   example-command arg-name1={'key':'value','obj'={'prop':"value"}}

Both JSON and Python formatting should work, including both styles of
string literal quotes. Both paradigms of literal values should work,
including null/true/false for JSON and None/True/False for Python.


Transactions have the following multi-line format:

   transaction(
   action-name1 [ arg-name1=arg1 ] ... [arg-nameN=argN ]
   ...
   action-nameN [ arg-name1=arg1 ] ... [arg-nameN=argN ]
   )

One line transactions are also supported:

   transaction( action-name1 ... )

For example:

    (QEMU) transaction(
    TRANS> block-dirty-bitmap-add node=drive0 name=bitmap1
    TRANS> block-dirty-bitmap-clear node=drive0 name=bitmap0
    TRANS> )
    {"return": {}}
    (QEMU)

Use the -v and -p options to activate the verbose and pretty-print options,
which will echo back the properly formatted JSON-compliant QMP that is being
sent to QEMU, which is useful for debugging and documentation generation.
"""

import argparse
import ast
import json
import logging
import os
import re
import readline
from subprocess import Popen
import sys
from typing import (
    IO,
    Dict,
    Iterator,
    List,
    NoReturn,
    Optional,
    Sequence,
    cast,
)

from . import (
    ConnectError,
    ExecuteError,
    QMPError,
    SocketAddrT,
)
from .legacy import (
    QEMUMonitorProtocol,
    QMPBadPortError,
    QMPMessage,
    QMPObject,
)


LOG = logging.getLogger(__name__)


class QMPCompleter:
    """
    QMPCompleter provides a readline library tab-complete behavior.
    """
    # NB: Python 3.9+ will probably allow us to subclass list[str] directly,
    # but pylint as of today does not know that List[str] is simply 'list'.
    def __init__(self) -> None:
        self._matches: List[str] = []

    def append(self, value: str) -> None:
        """Append a new valid completion to the list of possibilities."""
        return self._matches.append(value)

    def complete(self, text: str, state: int) -> Optional[str]:
        """readline.set_completer() callback implementation."""
        for cmd in self._matches:
            if cmd.startswith(text):
                if state == 0:
                    return cmd
                state -= 1
        return None


class QMPShellError(QMPError):
    """
    QMP Shell Base error class.
    """


class FuzzyJSON(ast.NodeTransformer):
    """
    This extension of ast.NodeTransformer filters literal "true/false/null"
    values in a Python AST and replaces them by proper "True/False/None" values
    that Python can properly evaluate.
    """

    @classmethod
    def visit_Name(cls,  # pylint: disable=invalid-name
                   node: ast.Name) -> ast.AST:
        """
        Transform Name nodes with certain values into Constant (keyword) nodes.
        """
        if node.id == 'true':
            return ast.Constant(value=True)
        if node.id == 'false':
            return ast.Constant(value=False)
        if node.id == 'null':
            return ast.Constant(value=None)
        return node


class QMPShell(QEMUMonitorProtocol):
    """
    QMPShell provides a basic readline-based QMP shell.

    :param address: Address of the QMP server.
    :param pretty: Pretty-print QMP messages.
    :param verbose: Echo outgoing QMP messages to console.
    """
    def __init__(self, address: SocketAddrT,
                 pretty: bool = False,
                 verbose: bool = False,
                 server: bool = False,
                 logfile: Optional[str] = None):
        super().__init__(address, server=server)
        self._greeting: Optional[QMPMessage] = None
        self._completer = QMPCompleter()
        self._transmode = False
        self._actions: List[QMPMessage] = []
        self._histfile = os.path.join(os.path.expanduser('~'),
                                      '.qmp-shell_history')
        self.pretty = pretty
        self.verbose = verbose
        self.logfile = None

        if logfile is not None:
            self.logfile = open(logfile, "w", encoding='utf-8')

    def close(self) -> None:
        # Hook into context manager of parent to save shell history.
        self._save_history()
        super().close()

    def _fill_completion(self) -> None:
        try:
            cmds = cast(List[Dict[str, str]], self.cmd('query-commands'))
            for cmd in cmds:
                self._completer.append(cmd['name'])
        except ExecuteError:
            pass

    def _completer_setup(self) -> None:
        self._completer = QMPCompleter()
        self._fill_completion()
        readline.set_history_length(1024)
        readline.set_completer(self._completer.complete)
        readline.parse_and_bind("tab: complete")
        # NB: default delimiters conflict with some command names
        # (eg. query-), clearing everything as it doesn't seem to matter
        readline.set_completer_delims('')
        try:
            readline.read_history_file(self._histfile)
        except FileNotFoundError:
            pass
        except IOError as err:
            msg = f"Failed to read history '{self._histfile}': {err!s}"
            LOG.warning(msg)

    def _save_history(self) -> None:
        try:
            readline.write_history_file(self._histfile)
        except IOError as err:
            msg = f"Failed to save history file '{self._histfile}': {err!s}"
            LOG.warning(msg)

    @classmethod
    def _parse_value(cls, val: str) -> object:
        try:
            return int(val)
        except ValueError:
            pass

        if val.lower() == 'true':
            return True
        if val.lower() == 'false':
            return False
        if val.startswith(('{', '[')):
            # Try first as pure JSON:
            try:
                return json.loads(val)
            except ValueError:
                pass
            # Try once again as FuzzyJSON:
            try:
                tree = ast.parse(val, mode='eval')
                transformed = FuzzyJSON().visit(tree)
                return ast.literal_eval(transformed)
            except (SyntaxError, ValueError):
                pass
        return val

    def _cli_expr(self,
                  tokens: Sequence[str],
                  parent: QMPObject) -> None:
        for arg in tokens:
            (key, sep, val) = arg.partition('=')
            if sep != '=':
                raise QMPShellError(
                    f"Expected a key=value pair, got '{arg!s}'"
                )

            value = self._parse_value(val)
            optpath = key.split('.')
            curpath = []
            for path in optpath[:-1]:
                curpath.append(path)
                obj = parent.get(path, {})
                if not isinstance(obj, dict):
                    msg = 'Cannot use "{:s}" as both leaf and non-leaf key'
                    raise QMPShellError(msg.format('.'.join(curpath)))
                parent[path] = obj
                parent = obj
            if optpath[-1] in parent:
                if isinstance(parent[optpath[-1]], dict):
                    msg = 'Cannot use "{:s}" as both leaf and non-leaf key'
                    raise QMPShellError(msg.format('.'.join(curpath)))
                raise QMPShellError(f'Cannot set "{key}" multiple times')
            parent[optpath[-1]] = value

    def _build_cmd(self, cmdline: str) -> Optional[QMPMessage]:
        """
        Build a QMP input object from a user provided command-line in the
        following format:

            < command-name > [ arg-name1=arg1 ] ... [ arg-nameN=argN ]
        """
        argument_regex = r'''(?:[^\s"']|"(?:\\.|[^"])*"|'(?:\\.|[^'])*')+'''
        cmdargs = re.findall(argument_regex, cmdline)
        qmpcmd: QMPMessage

        # Transactional CLI entry:
        if cmdargs and cmdargs[0] == 'transaction(':
            self._transmode = True
            self._actions = []
            cmdargs.pop(0)

        # Transactional CLI exit:
        if cmdargs and cmdargs[0] == ')' and self._transmode:
            self._transmode = False
            if len(cmdargs) > 1:
                msg = 'Unexpected input after close of Transaction sub-shell'
                raise QMPShellError(msg)
            qmpcmd = {
                'execute': 'transaction',
                'arguments': {'actions': self._actions}
            }
            return qmpcmd

        # No args, or no args remaining
        if not cmdargs:
            return None

        if self._transmode:
            # Parse and cache this Transactional Action
            finalize = False
            action = {'type': cmdargs[0], 'data': {}}
            if cmdargs[-1] == ')':
                cmdargs.pop(-1)
                finalize = True
            self._cli_expr(cmdargs[1:], action['data'])
            self._actions.append(action)
            return self._build_cmd(')') if finalize else None

        # Standard command: parse and return it to be executed.
        qmpcmd = {'execute': cmdargs[0], 'arguments': {}}
        self._cli_expr(cmdargs[1:], qmpcmd['arguments'])
        return qmpcmd

    def _print(self, qmp_message: object, fh: IO[str] = sys.stdout) -> None:
        jsobj = json.dumps(qmp_message,
                           indent=4 if self.pretty else None,
                           sort_keys=self.pretty)
        print(str(jsobj), file=fh)

    def _execute_cmd(self, cmdline: str) -> bool:
        try:
            qmpcmd = self._build_cmd(cmdline)
        except QMPShellError as err:
            print(
                f"Error while parsing command line: {err!s}\n"
                "command format: <command-name> "
                "[arg-name1=arg1] ... [arg-nameN=argN",
                file=sys.stderr
            )
            return True
        # For transaction mode, we may have just cached the action:
        if qmpcmd is None:
            return True
        if self.verbose:
            self._print(qmpcmd)
        resp = self.cmd_obj(qmpcmd)
        if resp is None:
            print('Disconnected')
            return False
        self._print(resp)
        if self.logfile is not None:
            cmd = {**qmpcmd, **resp}
            self._print(cmd, fh=self.logfile)
        return True

    def connect(self, negotiate: bool = True) -> None:
        self._greeting = super().connect(negotiate)
        self._completer_setup()

    def show_banner(self,
                    msg: str = 'Welcome to the QMP low-level shell!') -> None:
        """
        Print to stdio a greeting, and the QEMU version if available.
        """
        print(msg)
        if not self._greeting:
            print('Connected')
            return
        version = self._greeting['QMP']['version']['qemu']
        print("Connected to QEMU {major}.{minor}.{micro}\n".format(**version))

    @property
    def prompt(self) -> str:
        """
        Return the current shell prompt, including a trailing space.
        """
        if self._transmode:
            return 'TRANS> '
        return '(QEMU) '

    def read_exec_command(self) -> bool:
        """
        Read and execute a command.

        @return True if execution was ok, return False if disconnected.
        """
        try:
            cmdline = input(self.prompt)
        except EOFError:
            print()
            return False

        if cmdline == '':
            for event in self.get_events():
                print(event)
            return True

        return self._execute_cmd(cmdline)

    def repl(self) -> Iterator[None]:
        """
        Return an iterator that implements the REPL.
        """
        self.show_banner()
        while self.read_exec_command():
            yield
        self.close()


class HMPShell(QMPShell):
    """
    HMPShell provides a basic readline-based HMP shell, tunnelled via QMP.

    :param address: Address of the QMP server.
    :param pretty: Pretty-print QMP messages.
    :param verbose: Echo outgoing QMP messages to console.
    """
    def __init__(self, address: SocketAddrT,
                 pretty: bool = False,
                 verbose: bool = False,
                 server: bool = False,
                 logfile: Optional[str] = None):
        super().__init__(address, pretty, verbose, server, logfile)
        self._cpu_index = 0

    def _cmd_completion(self) -> None:
        for cmd in self._cmd_passthrough('help')['return'].split('\r\n'):
            if cmd and cmd[0] != '[' and cmd[0] != '\t':
                name = cmd.split()[0]  # drop help text
                if name == 'info':
                    continue
                if name.find('|') != -1:
                    # Command in the form 'foobar|f' or 'f|foobar', take the
                    # full name
                    opt = name.split('|')
                    if len(opt[0]) == 1:
                        name = opt[1]
                    else:
                        name = opt[0]
                self._completer.append(name)
                self._completer.append('help ' + name)  # help completion

    def _info_completion(self) -> None:
        for cmd in self._cmd_passthrough('info')['return'].split('\r\n'):
            if cmd:
                self._completer.append('info ' + cmd.split()[1])

    def _other_completion(self) -> None:
        # special cases
        self._completer.append('help info')

    def _fill_completion(self) -> None:
        self._cmd_completion()
        self._info_completion()
        self._other_completion()

    def _cmd_passthrough(self, cmdline: str,
                         cpu_index: int = 0) -> QMPMessage:
        return self.cmd_obj({
            'execute': 'human-monitor-command',
            'arguments': {
                'command-line': cmdline,
                'cpu-index': cpu_index
            }
        })

    def _execute_cmd(self, cmdline: str) -> bool:
        if cmdline.split()[0] == "cpu":
            # trap the cpu command, it requires special setting
            try:
                idx = int(cmdline.split()[1])
                if 'return' not in self._cmd_passthrough('info version', idx):
                    print('bad CPU index')
                    return True
                self._cpu_index = idx
            except ValueError:
                print('cpu command takes an integer argument')
                return True
        resp = self._cmd_passthrough(cmdline, self._cpu_index)
        if resp is None:
            print('Disconnected')
            return False
        assert 'return' in resp or 'error' in resp
        if 'return' in resp:
            # Success
            if len(resp['return']) > 0:
                print(resp['return'], end=' ')
        else:
            # Error
            print('%s: %s' % (resp['error']['class'], resp['error']['desc']))
        return True

    def show_banner(self, msg: str = 'Welcome to the HMP shell!') -> None:
        QMPShell.show_banner(self, msg)


def die(msg: str) -> NoReturn:
    """Write an error to stderr, then exit with a return code of 1."""
    sys.stderr.write('ERROR: %s\n' % msg)
    sys.exit(1)


def main() -> None:
    """
    qmp-shell entry point: parse command line arguments and start the REPL.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('-H', '--hmp', action='store_true',
                        help='Use HMP interface')
    parser.add_argument('-N', '--skip-negotiation', action='store_true',
                        help='Skip negotiate (for qemu-ga)')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Verbose (echo commands sent and received)')
    parser.add_argument('-p', '--pretty', action='store_true',
                        help='Pretty-print JSON')
    parser.add_argument('-l', '--logfile',
                        help='Save log of all QMP messages to PATH')

    default_server = os.environ.get('QMP_SOCKET')
    parser.add_argument('qmp_server', action='store',
                        default=default_server,
                        help='< UNIX socket path | TCP address:port >')

    args = parser.parse_args()
    if args.qmp_server is None:
        parser.error("QMP socket or TCP address must be specified")

    shell_class = HMPShell if args.hmp else QMPShell

    try:
        address = shell_class.parse_address(args.qmp_server)
    except QMPBadPortError:
        parser.error(f"Bad port number: {args.qmp_server}")
        return  # pycharm doesn't know error() is noreturn

    with shell_class(address, args.pretty, args.verbose, args.logfile) as qemu:
        try:
            qemu.connect(negotiate=not args.skip_negotiation)
        except ConnectError as err:
            if isinstance(err.exc, OSError):
                die(f"Couldn't connect to {args.qmp_server}: {err!s}")
            die(str(err))

        for _ in qemu.repl():
            pass


def main_wrap() -> None:
    """
    qmp-shell-wrap entry point: parse command line arguments and
    start the REPL.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('-H', '--hmp', action='store_true',
                        help='Use HMP interface')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='Verbose (echo commands sent and received)')
    parser.add_argument('-p', '--pretty', action='store_true',
                        help='Pretty-print JSON')
    parser.add_argument('-l', '--logfile',
                        help='Save log of all QMP messages to PATH')

    parser.add_argument('command', nargs=argparse.REMAINDER,
                        help='QEMU command line to invoke')

    args = parser.parse_args()

    cmd = args.command
    if len(cmd) != 0 and cmd[0] == '--':
        cmd = cmd[1:]
    if len(cmd) == 0:
        cmd = ["qemu-system-x86_64"]

    sockpath = "qmp-shell-wrap-%d" % os.getpid()
    cmd += ["-qmp", "unix:%s" % sockpath]

    shell_class = HMPShell if args.hmp else QMPShell

    try:
        address = shell_class.parse_address(sockpath)
    except QMPBadPortError:
        parser.error(f"Bad port number: {sockpath}")
        return  # pycharm doesn't know error() is noreturn

    try:
        with shell_class(address, args.pretty, args.verbose,
                         True, args.logfile) as qemu:
            with Popen(cmd):

                try:
                    qemu.accept()
                except ConnectError as err:
                    if isinstance(err.exc, OSError):
                        die(f"Couldn't connect to {args.qmp_server}: {err!s}")
                    die(str(err))

                for _ in qemu.repl():
                    pass
    finally:
        os.unlink(sockpath)


if __name__ == '__main__':
    main()



"""
VM Lifecycle Manager

Manages VM lifecycle with transactional storage operations.
Provides atomic operations for VM creation, deletion, and storage management
to prevent orphaned storage and inconsistent state between database and filesystem.

Part of the Transactional Storage Management feature (Phase 2).
"""

from pathlib import Path
from typing import Dict, List

from .exceptions import StateError
from .logger import LOG
from .state import StateManager
from .storage_registry import StorageRegistry


class VMLifecycleManager:
    """
    Manages VM lifecycle with transactional storage.

    Provides atomic operations that ensure VM entries and storage files
    are always in sync. If any operation fails, all changes are rolled back
    to maintain consistency.

    Key features:
    - Atomic VM + storage creation
    - Explicit storage control on deletion
    - Orphaned storage reattachment
    - Transactional safety with rollback

    Integrates with:
    - StateManager for VM database operations
    - StorageRegistry for storage tracking

    Note: This implementation uses a compensating transaction pattern where
    if storage registration fails after VM creation, the VM is deleted to
    maintain consistency. This is simpler than trying to wrap StateManager's
    internal connection management.
    """

    def __init__(self, state_manager: StateManager):
        """
        Initialize VM lifecycle manager.

        Args:
            state_manager: StateManager instance for VM database operations
        """
        self.state_manager = state_manager

        # Create a separate connection for storage registry operations
        # This is safe because storage registry only reads/writes the storage_registry table
        # while StateManager operations work on vm_instances table
        self._db_conn = self.state_manager._create_connection()

        # Initialize storage registry with its own connection and data dir
        self.storage_registry = StorageRegistry(
            self._db_conn, data_dir=self.state_manager.xdg.data_dir
        )

    def create_vm_with_storage(
        self,
        vm_name: str,
        vm_config: Dict,
        disk_paths: List[Path],
    ) -> str:
        """
        Create VM and register its storage atomically.

        This method ensures that either both the VM entry and storage registrations
        succeed, or neither happens (atomic operation). If any disk path doesn't exist
        or registration fails, the entire operation is rolled back.

        Uses a compensating transaction pattern: if storage registration fails after
        VM creation, the VM is deleted to maintain consistency.

        Args:
            vm_name: VM name (must be unique)
            vm_config: VM configuration dictionary
            disk_paths: List of disk image paths to register with the VM

        Returns:
            VM instance ID

        Raises:
            ValueError: If any disk path doesn't exist
            StateError: If VM creation fails (name conflict, etc.)
            Exception: If database operation fails
        """
        # Validate all disk paths exist BEFORE starting operations
        for disk_path in disk_paths:
            if not disk_path.exists():
                raise ValueError(f"Disk image not found: {disk_path}")

        vm_id = None
        try:
            # Step 1: Create VM entry in database
            vm_id = self.state_manager.create_vm(vm_name, vm_config)
            LOG.debug(f"Created VM '{vm_name}' with ID {vm_id}")

            # Step 2: Register all storage files
            for disk_path in disk_paths:
                self.storage_registry.register_storage(
                    vm_name,
                    disk_path,
                    storage_type="disk",
                )
                LOG.debug(f"Registered storage for VM '{vm_name}': {disk_path}")

            LOG.info(
                f"Successfully created VM '{vm_name}' with {len(disk_paths)} storage file(s)"
            )
            return vm_id

        except Exception as e:
            # Compensating transaction: if storage registration failed after VM creation,
            # delete the VM to maintain consistency
            if vm_id is not None:
                LOG.warning(
                    f"Storage registration failed for VM '{vm_name}', rolling back VM creation"
                )
                try:
                    self.state_manager.remove_vm(vm_name)
                    LOG.debug(f"Rolled back VM creation: '{vm_name}'")
                except Exception as rollback_error:
                    LOG.error(
                        f"Failed to rollback VM creation for '{vm_name}': {rollback_error}"
                    )
            raise

    def delete_vm(
        self,
        vm_name: str,
        delete_storage: bool = False,
        keep_snapshots: bool = False,
    ) -> None:
        """
        Delete VM with explicit storage handling.

        Provides fine-grained control over what happens to storage files when
        a VM is deleted. By default, storage is kept to prevent accidental data loss.

        Safety checks:
        - Verifies VM is stopped before deletion
        - Atomic operation (VM + storage registry changes via CASCADE DELETE)
        - Optional storage file deletion with snapshot preservation

        Args:
            vm_name: VM name to delete
            delete_storage: If True, delete storage files from disk (default: False)
            keep_snapshots: If True, keep snapshot files even if delete_storage=True

        Raises:
            ValueError: If VM is running (must stop first for safety)
            StateError: If VM doesn't exist or database operation fails
        """
        # Get VM instance to check status
        vm_instance = self.state_manager.get_vm(vm_name)
        if not vm_instance:
            raise StateError(f"VM '{vm_name}' not found")

        # Safety check: Verify VM is stopped
        if vm_instance.status == "running":
            raise ValueError(
                f"Cannot delete running VM '{vm_name}'. Stop it first with: maqet stop {vm_name}"
            )

        # Get all storage for VM BEFORE deletion (CASCADE will remove registry entries)
        storage_entries = self.storage_registry.get_vm_storage(vm_name)
        LOG.debug(f"Found {len(storage_entries)} storage entries for VM '{vm_name}'")

        # Delete VM entry (CASCADE DELETE will automatically remove storage registry entries)
        success = self.state_manager.remove_vm(vm_name)
        if not success:
            raise StateError(f"Failed to delete VM '{vm_name}' from database")

        LOG.debug(f"Deleted VM '{vm_name}' from database")

        # Handle storage files if requested
        if delete_storage:
            deleted_count = 0
            skipped_count = 0

            for entry in storage_entries:
                # Skip snapshots if keep_snapshots is True
                if keep_snapshots and entry.storage_type == "snapshot":
                    LOG.debug(f"Keeping snapshot: {entry.storage_path}")
                    skipped_count += 1
                    continue

                # Delete the file if it exists
                if entry.storage_path.exists():
                    try:
                        entry.storage_path.unlink()
                        LOG.debug(f"Deleted storage file: {entry.storage_path}")
                        deleted_count += 1
                    except OSError as e:
                        LOG.warning(
                            f"Failed to delete storage file {entry.storage_path}: {e}"
                        )
                else:
                    LOG.debug(f"Storage file already gone: {entry.storage_path}")

            LOG.info(
                f"VM '{vm_name}' deleted with {deleted_count} storage file(s) removed"
                + (f", {skipped_count} snapshot(s) kept" if skipped_count > 0 else "")
            )
        else:
            LOG.info(
                f"VM '{vm_name}' deleted, {len(storage_entries)} storage file(s) kept"
            )

    def attach_orphaned_storage(
        self,
        vm_name: str,
        storage_path: Path,
        vm_config: Dict,
    ) -> str:
        """
        Create VM entry for existing orphaned storage.

        Allows reattaching storage files that exist on disk but have no VM entry.
        This is useful for recovering from manual deletions or database issues.

        Uses the same compensating transaction pattern as create_vm_with_storage.

        Args:
            vm_name: New VM name to create
            storage_path: Path to orphaned storage file
            vm_config: VM configuration dictionary

        Returns:
            VM instance ID

        Raises:
            ValueError: If storage doesn't exist or VM name already exists
            StateError: If VM creation fails
        """
        # Validate storage exists
        if not storage_path.exists():
            raise ValueError(f"Storage file not found: {storage_path}")

        # Validate VM doesn't already exist
        existing_vm = self.state_manager.get_vm(vm_name)
        if existing_vm:
            raise ValueError(
                f"VM '{vm_name}' already exists. "
                f"Use a different name or delete the existing VM first."
            )

        vm_id = None
        try:
            # Create VM entry
            vm_id = self.state_manager.create_vm(vm_name, vm_config)
            LOG.debug(f"Created VM '{vm_name}' for orphaned storage")

            # Register existing storage
            self.storage_registry.register_storage(
                vm_name,
                storage_path,
                storage_type="disk",
            )

            LOG.info(
                f"Successfully attached orphaned storage to VM '{vm_name}': {storage_path}"
            )
            return vm_id

        except Exception as e:
            # Compensating transaction: rollback VM creation if storage registration failed
            if vm_id is not None:
                LOG.warning(
                    f"Storage registration failed for VM '{vm_name}', rolling back VM creation"
                )
                try:
                    self.state_manager.remove_vm(vm_name)
                    LOG.debug(f"Rolled back VM creation: '{vm_name}'")
                except Exception as rollback_error:
                    LOG.error(
                        f"Failed to rollback VM creation for '{vm_name}': {rollback_error}"
                    )
            raise

    def close(self) -> None:
        """
        Close database connection used by storage registry.

        Purpose: Clean up database resources to prevent ResourceWarnings.
        This should be called when the VMLifecycleManager is no longer needed.

        Note: The connection is created in __init__ and stored in self._db_conn.
        """
        if hasattr(self, '_db_conn') and self._db_conn:
            try:
                self._db_conn.close()
                LOG.debug("VMLifecycleManager database connection closed")
            except Exception as e:
                LOG.debug(f"Error closing VMLifecycleManager connection: {e}")

    def __del__(self) -> None:
        """
        Destructor to ensure database connection is closed.

        Purpose: Safety net for resource cleanup during garbage collection.
        Prefer explicit close() calls over relying on __del__.
        """
        try:
            self.close()
        except Exception:
            # Ignore errors during cleanup
            pass



"""
Advanced logging setup for MAQET with configurable verbosity and file output.

Provides colored console output, file logging, and verbosity level control.
"""

import logging
import sys
from pathlib import Path
from typing import Optional


class ColoredFormatter(logging.Formatter):
    """Custom formatter with colored output for different log levels."""

    # Bold colors for severe levels
    RED_BOLD = "\x1b[31;1m"
    YELLOW_BOLD = "\x1b[33;1m"
    CYAN_BOLD = "\x1b[36;1m"

    # Dim/regular colors for info levels
    GREY_DIM = "\x1b[90m"  # Dim grey for debug
    WHITE = "\x1b[97m"  # Bright white for info

    # Reset
    RESET = "\x1b[0m"

    MESSAGE_FORMAT = " %(message)s "

    FORMATS = {
        logging.DEBUG: GREY_DIM
        + "D"
        + RESET
        + GREY_DIM
        + MESSAGE_FORMAT
        + RESET,
        logging.INFO: WHITE + "I" + RESET + WHITE + MESSAGE_FORMAT + RESET,
        logging.WARNING: CYAN_BOLD + "W" + RESET + MESSAGE_FORMAT,
        logging.ERROR: YELLOW_BOLD + "E" + RESET + MESSAGE_FORMAT,
        logging.CRITICAL: RED_BOLD + "C" + RESET + MESSAGE_FORMAT,
    }

    def format(self, record):
        """Format log record with appropriate color."""
        log_fmt = self.FORMATS.get(record.levelno, self.MESSAGE_FORMAT)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class MaqetLogger:
    """MAQET logging manager with configurable verbosity and file output.

    Features:
    - Colored console output for improved readability
    - Rotating file logs (10MB max, 5 backups = 60MB total)
    - Configurable verbosity levels (ERROR/WARNING/INFO/DEBUG)
    """

    def __init__(self, name: str = "maqet"):
        """
        Initialize the logger.

        Args:
            name: Logger name
        """
        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.DEBUG)
        self.logger.propagate = False

        # Clear any existing handlers
        self.logger.handlers.clear()

        # Setup console handler with colored formatting
        self._setup_console_handler()

        # File handler will be added when configure_file_logging is called
        self._file_handler: Optional[logging.FileHandler] = None

    def _setup_console_handler(self):
        """Setup colored console output handler."""
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setFormatter(ColoredFormatter())
        console_handler.setLevel(logging.ERROR)  # Default console level
        self.logger.addHandler(console_handler)

    def set_verbosity(self, verbosity_level: int):
        """
        Set logging verbosity level.

        Args:
            verbosity_level: 0=ERROR, 1=WARNING, 2=INFO, 3+=DEBUG
        """
        level_map = {
            0: logging.ERROR,    # 40 - No -v flags (default, errors only)
            1: logging.WARNING,  # 30 - -v (shows warnings + errors)
            2: logging.INFO,     # 20 - -vv
            3: logging.DEBUG,    # 10 - -vvv or more
        }

        # Cap at DEBUG level for anything >= 3
        if verbosity_level >= 3:
            level = logging.DEBUG
        else:
            level = level_map.get(verbosity_level, logging.ERROR)

        # Update console handler level
        for handler in self.logger.handlers:
            if (
                isinstance(handler, logging.StreamHandler)
                and handler.stream == sys.stdout
            ):
                handler.setLevel(level)
                break

    def configure_file_logging(
        self, log_file: Optional[Path] = None, mode: str = "w"
    ):
        """
        Configure file logging output with rotation.

        Uses RotatingFileHandler to prevent unbounded log growth:
        - Maximum file size: 10MB
        - Backup count: 5 files
        - Total maximum size: 60MB (current + 5 backups)

        Args:
            log_file: Path to log file. If None, uses 'maqet.log' in current directory
            mode: File open mode ('w' for overwrite, 'a' for append) - ignored for rotating handler
        """
        from logging.handlers import RotatingFileHandler

        if self._file_handler:
            self.logger.removeHandler(self._file_handler)
            self._file_handler = None

        if log_file is None:
            log_file = Path("maqet.log")

        # Create log file parent directory if it doesn't exist
        log_file.parent.mkdir(parents=True, exist_ok=True)

        # Setup rotating file handler with 10MB max size and 5 backups
        max_bytes = 10 * 1024 * 1024  # 10MB
        backup_count = 5  # Keep 5 old log files

        self._file_handler = RotatingFileHandler(
            log_file, maxBytes=max_bytes, backupCount=backup_count
        )
        self._file_handler.setLevel(
            logging.DEBUG
        )  # Always log everything to file

        file_formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        self._file_handler.setFormatter(file_formatter)

        self.logger.addHandler(self._file_handler)

    def get_logger(self) -> logging.Logger:
        """Get the underlying logger instance."""
        return self.logger


# Global logger instance
_maqet_logger = MaqetLogger()
LOG = _maqet_logger.get_logger()


def set_verbosity(level: int):
    """Set global logging verbosity level."""
    _maqet_logger.set_verbosity(level)


def configure_file_logging(log_file: Optional[Path] = None, mode: str = "w"):
    """Configure global file logging."""
    _maqet_logger.configure_file_logging(log_file, mode)



"""
Process Lifecycle Manager

Handles QEMU process lifecycle operations including:
- Graceful shutdown via QMP
- Force kill with SIGTERM/SIGKILL escalation
- Process state tracking
- Cleanup coordination after VM stop
- Instance-scoped PID registry management

Extracted from Machine class to follow Single Responsibility Principle.

CRITICAL ARCHITECTURE: Each ProcessLifecycleManager instance maintains its own
PID registry to prevent cross-contamination between Maqet instances in concurrent
test environments. This ensures one instance's cleanup never affects another's.
"""

import atexit
import os
import signal
import threading
import time
from typing import TYPE_CHECKING, Dict, Optional, Set

from ..constants import Intervals, Timeouts
from ..logger import LOG
from ..utils.process_validation import validate_pid_exists

if TYPE_CHECKING:
    from qemu.machine import QEMUMachine
    from ..state import StateManager


class ProcessLifecycleManager:
    """
    Manages QEMU process lifecycle operations.

    Responsibilities:
    - Graceful shutdown via QMP
    - Force kill with signal escalation
    - Process state checking
    - Cleanup coordination (PID registry, files, database state)

    CRITICAL: Each instance maintains its own process registry to prevent
    cross-contamination between Maqet instances in concurrent tests.
    """

    def __init__(
        self,
        vm_id: str,
        state_manager: "StateManager",
    ):
        """
        Initialize ProcessLifecycleManager.

        Args:
            vm_id: VM identifier
            state_manager: State manager instance for file paths and status updates
        """
        self.vm_id = vm_id
        self.state_manager = state_manager

        # Instance-level registry of active QEMU PIDs for cleanup on exit
        # Maps PID to VM ID for better tracking and debugging
        self._qemu_process_registry: Dict[int, str] = {}
        self._registry_lock = threading.Lock()

        # Register instance cleanup on exit
        atexit.register(self.cleanup_all_qemu_processes)

    def graceful_shutdown(
        self,
        qemu_machine: "QEMUMachine",
        timeout: int = Timeouts.VM_GRACEFUL_SHUTDOWN,
    ) -> bool:
        """
        Attempt graceful shutdown of VM via QMP.

        Args:
            qemu_machine: QEMU machine instance
            timeout: Maximum seconds to wait for shutdown

        Returns:
            True if shutdown succeeded, False otherwise
        """
        if not qemu_machine:
            return False

        try:
            LOG.debug(f"Attempting graceful shutdown of VM {self.vm_id}")
            qemu_machine.shutdown()

            # Wait for process to exit
            start_time = time.time()
            while (
                qemu_machine.is_running()
                and (time.time() - start_time) < timeout
            ):
                time.sleep(Intervals.SHUTDOWN_POLL)

            # Check if shutdown succeeded
            if not qemu_machine.is_running():
                LOG.info(f"VM {self.vm_id} shutdown gracefully")
                return True
            else:
                LOG.warning(
                    f"VM {self.vm_id} didn't shutdown gracefully in {timeout}s"
                )
                return False

        except Exception as e:
            LOG.warning(f"Graceful shutdown failed for VM {self.vm_id}: {e}")
            return False

    def force_kill(self, pid: int) -> bool:
        """
        Force kill the VM process using SIGTERM then SIGKILL.

        Sends SIGTERM first, waits briefly, then sends SIGKILL if needed.
        Verifies the process is actually dead after kill attempts.

        This method NEVER raises exceptions - it only logs errors and returns
        a status boolean. This ensures cleanup paths (like __del__) are never
        disrupted.

        Args:
            pid: Process ID to kill

        Returns:
            True if kill succeeded (or process already dead), False if kill failed
        """
        if not pid:
            return True

        LOG.info(f"Force killing VM {self.vm_id} (PID {pid})")
        try:
            # Send SIGTERM first
            os.kill(pid, signal.SIGTERM)
            time.sleep(Intervals.SIGTERM_WAIT)

            # Check if still alive, send SIGKILL if needed
            if self.is_process_alive(pid):
                LOG.debug(f"Process {pid} still alive, sending SIGKILL")
                os.kill(pid, signal.SIGKILL)
                time.sleep(0.2)  # Brief wait for SIGKILL to take effect
        except ProcessLookupError:
            # Process already dead - success
            return True

        # Final verification that process is actually dead
        if self.is_process_alive(pid):
            LOG.critical(f"Failed to kill QEMU process {pid} for VM {self.vm_id} after SIGTERM and SIGKILL")
            return False

        return True

    def is_process_alive(self, pid: int) -> bool:
        """
        Check if process is still running.

        Args:
            pid: Process ID to check

        Returns:
            True if process is alive, False otherwise
        """
        return validate_pid_exists(pid, process_name="process", log_permission_warning=False)

    def cleanup_after_stop(self, pid: Optional[int]) -> None:
        """
        Cleanup after VM stops.

        GUARANTEED to run in finally block. NEVER raises exceptions - only logs errors.

        - Unregisters PID from active registry
        - Updates database status
        - Removes temporary files

        Args:
            pid: Process ID that was stopped (None if already cleaned up)
        """
        try:
            # Unregister PID from cleanup registry
            if pid:
                with self._registry_lock:
                    self._qemu_process_registry.pop(pid, None)
        except Exception as e:
            LOG.warning(f"Failed to unregister PID {pid}: {e}")

        try:
            # Update database status to stopped
            self.state_manager.update_vm_status(
                self.vm_id, "stopped", pid=None, socket_path=None, qmp_socket_path=None
            )
        except Exception as e:
            LOG.warning(f"Failed to update VM status for {self.vm_id}: {e}")

        try:
            # Clean up files
            self.cleanup_files()
        except Exception as e:
            LOG.warning(f"Failed to cleanup files for {self.vm_id}: {e}")

    def cleanup_files(self) -> None:
        """Clean up temporary files (PID file, socket handled by QEMUMachine)."""
        # Remove PID file
        pid_path = self.state_manager.get_pid_path(self.vm_id)
        if pid_path.exists():
            pid_path.unlink()

        # Socket cleanup is handled by QEMUMachine

    def register_pid(self, pid: int) -> None:
        """
        Register PID in instance cleanup registry (thread-safe).

        Args:
            pid: Process ID to register
        """
        with self._registry_lock:
            self._qemu_process_registry[pid] = self.vm_id

    def unregister_pid(self, pid: int) -> None:
        """
        Unregister PID from instance cleanup registry (thread-safe).

        Args:
            pid: Process ID to unregister
        """
        with self._registry_lock:
            self._qemu_process_registry.pop(pid, None)

    def emergency_cleanup(self, pid: int) -> None:
        """
        Emergency cleanup if normal cleanup didn't happen.

        This is called by atexit handler to ensure process is killed even if
        normal stop() wasn't called.

        Args:
            pid: Process ID to cleanup
        """
        if not pid:
            return

        if self.is_process_alive(pid):
            try:
                LOG.warning(f"Emergency cleanup: killing orphan QEMU process {pid} for VM {self.vm_id}")
                os.kill(pid, signal.SIGKILL)
            except ProcessLookupError:
                pass
            finally:
                with self._registry_lock:
                    self._qemu_process_registry.pop(pid, None)

    def get_active_pids(self) -> Set[int]:
        """
        Get set of all active QEMU PIDs for this instance.

        Returns:
            Set of active process IDs
        """
        with self._registry_lock:
            return set(self._qemu_process_registry.keys())

    def get_process_registry(self) -> Dict[int, str]:
        """
        Get copy of process registry for this instance (thread-safe).

        Returns:
            Dict mapping PID to VM ID
        """
        with self._registry_lock:
            return self._qemu_process_registry.copy()

    def cleanup_orphan_processes(self) -> None:
        """
        Kill any QEMU processes that are still running for this instance.

        This is typically called during Python exit to prevent orphaned processes.
        Thread-safe implementation using the instance registry lock.
        """
        with self._registry_lock:
            pids = list(self._qemu_process_registry.keys())

        if pids:
            LOG.debug(
                f"Cleaning up {len(pids)} orphan QEMU processes for instance"
            )
            for pid in pids:
                try:
                    os.kill(pid, signal.SIGKILL)
                    LOG.debug(f"Killed orphan QEMU process {pid}")
                except (ProcessLookupError, OSError):
                    pass  # Process already dead

            with self._registry_lock:
                self._qemu_process_registry.clear()

    def cleanup_all_qemu_processes(self) -> None:
        """
        Cleanup all tracked QEMU processes for this instance.

        Instance method registered with atexit to ensure all QEMU processes
        managed by this instance are cleaned up on Python exit.
        Thread-safe implementation using the instance registry lock.

        This method can be called from tests to ensure all QEMU processes
        for a specific Maqet instance are cleaned up between test runs.
        """
        with self._registry_lock:
            pids = list(self._qemu_process_registry.keys())

        for pid in pids:
            try:
                os.kill(pid, signal.SIGKILL)
                LOG.debug(f"Killed QEMU process {pid}")
            except (ProcessLookupError, OSError):
                pass  # Process already dead

        with self._registry_lock:
            self._qemu_process_registry.clear()



"""Manager classes for Maqet components."""

from .config_manager import ConfigManager
from .process_lifecycle_manager import ProcessLifecycleManager
from .qmp_manager import QMPManager
from .vm_manager import VMManager

__all__ = [
    "ConfigManager",
    "VMManager",
    "QMPManager",
    "ProcessLifecycleManager",
]



"""
IPC Retry Logic with Exponential Backoff

Provides decorators and utilities for retrying IPC operations that may fail
due to transient network issues, temporary unavailability, or timing issues.

Features:
- Exponential backoff with configurable base delay
- Configurable max attempts
- Selective exception handling (retry only on specific exceptions)
- Debug logging for retry attempts
- Synchronous and asynchronous decorator support
"""

import asyncio
import functools
import time
from typing import Callable, Tuple, Type

from ..logger import LOG


def retry_with_backoff(
    max_attempts: int = 3,
    backoff_base: float = 0.5,
    exceptions: Tuple[Type[Exception], ...] = (
        ConnectionRefusedError,
        FileNotFoundError,
    ),
):
    """
    Decorator for synchronous functions with exponential backoff retry logic.

    Retries function on specified exceptions using exponential backoff:
    - Attempt 1: Immediate
    - Attempt 2: Wait backoff_base seconds (0.5s default)
    - Attempt 3: Wait backoff_base * 2 seconds (1s default)
    - Attempt 4: Wait backoff_base * 4 seconds (2s default)
    - etc.

    Args:
        max_attempts: Maximum number of attempts (default: 3)
        backoff_base: Base delay in seconds for exponential backoff (default: 0.5)
        exceptions: Tuple of exception types to retry on
                   (default: ConnectionRefusedError, FileNotFoundError)

    Returns:
        Decorated function that retries on transient failures

    Example:
        @retry_with_backoff(max_attempts=3, backoff_base=0.5)
        def connect_to_service():
            return service.connect()

    Note:
        - Only retries on specified exceptions (transient errors)
        - Other exceptions propagate immediately (permanent errors)
        - Logs retry attempts at DEBUG level
        - Final failure logs at WARNING level
    """

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            last_exception = None

            for attempt in range(1, max_attempts + 1):
                try:
                    return func(*args, **kwargs)

                except exceptions as e:
                    last_exception = e

                    if attempt < max_attempts:
                        # Calculate delay with exponential backoff
                        delay = backoff_base * (2 ** (attempt - 1))
                        LOG.debug(
                            f"IPC retry attempt {attempt}/{max_attempts} failed: {e}. "
                            f"Retrying in {delay:.2f}s..."
                        )
                        time.sleep(delay)
                    else:
                        # Max attempts reached
                        LOG.warning(
                            f"IPC operation failed after {max_attempts} attempts: {e}"
                        )

                except Exception as e:
                    # Non-retryable exception, propagate immediately
                    LOG.debug(f"Non-retryable exception in IPC operation: {e}")
                    raise

            # If we get here, all attempts failed
            raise last_exception

        return wrapper

    return decorator


def async_retry_with_backoff(
    max_attempts: int = 3,
    backoff_base: float = 0.5,
    exceptions: Tuple[Type[Exception], ...] = (
        ConnectionRefusedError,
        FileNotFoundError,
        OSError,
    ),
):
    """
    Decorator for asynchronous functions with exponential backoff retry logic.

    Async version of retry_with_backoff(). Retries async function on specified
    exceptions using exponential backoff with asyncio.sleep.

    Args:
        max_attempts: Maximum number of attempts (default: 3)
        backoff_base: Base delay in seconds for exponential backoff (default: 0.5)
        exceptions: Tuple of exception types to retry on
                   (default: ConnectionRefusedError, FileNotFoundError, OSError)

    Returns:
        Decorated async function that retries on transient failures

    Example:
        @async_retry_with_backoff(max_attempts=3, backoff_base=0.5)
        async def connect_to_socket():
            return await asyncio.open_unix_connection("/path/to/socket")

    Note:
        - Only retries on specified exceptions (transient errors)
        - Other exceptions propagate immediately (permanent errors)
        - Logs retry attempts at DEBUG level
        - Final failure logs at WARNING level
        - Uses asyncio.sleep for non-blocking delays
    """

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            last_exception = None

            for attempt in range(1, max_attempts + 1):
                try:
                    return await func(*args, **kwargs)

                except exceptions as e:
                    last_exception = e

                    if attempt < max_attempts:
                        # Calculate delay with exponential backoff
                        delay = backoff_base * (2 ** (attempt - 1))
                        LOG.debug(
                            f"IPC retry attempt {attempt}/{max_attempts} failed: {e}. "
                            f"Retrying in {delay:.2f}s..."
                        )
                        await asyncio.sleep(delay)
                    else:
                        # Max attempts reached
                        LOG.warning(
                            f"IPC operation failed after {max_attempts} attempts: {e}"
                        )

                except Exception as e:
                    # Non-retryable exception, propagate immediately
                    LOG.debug(f"Non-retryable exception in IPC operation: {e}")
                    raise

            # If we get here, all attempts failed
            raise last_exception

        return wrapper

    return decorator


class CircuitBreaker:
    """
    Circuit breaker pattern for IPC operations.

    Prevents repeated attempts to connect to a failing service by
    "opening the circuit" after a threshold of failures.

    States:
    - CLOSED: Normal operation, requests pass through
    - OPEN: Too many failures, requests fail immediately
    - HALF_OPEN: Testing if service recovered, allow one request

    Usage:
        breaker = CircuitBreaker(failure_threshold=5, timeout=60)

        if breaker.is_open():
            raise Exception("Circuit breaker open")

        try:
            result = connect_to_service()
            breaker.record_success()
        except Exception:
            breaker.record_failure()
            raise

    Note:
        - Circuit opens after failure_threshold consecutive failures
        - After timeout seconds, circuit enters HALF_OPEN state
        - One successful request in HALF_OPEN state closes circuit
        - Thread-safe for concurrent access
    """

    def __init__(self, failure_threshold: int = 5, timeout: float = 60.0):
        """
        Initialize circuit breaker.

        Args:
            failure_threshold: Number of failures before opening circuit
            timeout: Seconds to wait before attempting recovery (HALF_OPEN state)
        """
        self.failure_threshold = failure_threshold
        self.timeout = timeout
        self.failure_count = 0
        self.last_failure_time = None
        self.state = "CLOSED"  # CLOSED, OPEN, HALF_OPEN

    def is_open(self) -> bool:
        """
        Check if circuit is open (blocking requests).

        Returns:
            True if circuit is open and requests should be blocked
        """
        if self.state == "OPEN":
            # Check if timeout elapsed, transition to HALF_OPEN
            if (
                self.last_failure_time
                and time.time() - self.last_failure_time >= self.timeout
            ):
                LOG.debug("Circuit breaker entering HALF_OPEN state")
                self.state = "HALF_OPEN"
                return False
            return True
        return False

    def record_success(self) -> None:
        """Record successful operation, reset failure count."""
        if self.state == "HALF_OPEN":
            LOG.debug("Circuit breaker closing (recovered)")
        self.failure_count = 0
        self.state = "CLOSED"

    def record_failure(self) -> None:
        """Record failed operation, potentially open circuit."""
        self.failure_count += 1
        self.last_failure_time = time.time()

        if self.failure_count >= self.failure_threshold:
            LOG.warning(
                f"Circuit breaker opening after {self.failure_count} failures"
            )
            self.state = "OPEN"

    def reset(self) -> None:
        """Manually reset circuit breaker to CLOSED state."""
        LOG.debug("Circuit breaker manually reset")
        self.failure_count = 0
        self.last_failure_time = None
        self.state = "CLOSED"



"""
QMP Command Wrappers

Pre-built QMP command implementations that can be reused across the codebase.
These provide a clean abstraction layer over raw QMP commands.
"""

from typing import Any, Dict


class QMPCommands:
    """Collection of pre-built QMP command wrappers."""

    @staticmethod
    def screendump(machine, filename: str) -> Dict[str, Any]:
        """
        Take a screenshot of the VM display.

        Args:
            machine: Machine instance with qmp_command method
            filename: Path to save screenshot (PPM format)

        Returns:
            QMP command result
        """
        return machine.qmp_command("screendump", filename=filename)

    @staticmethod
    def pause(machine) -> Dict[str, Any]:
        """
        Pause VM execution.

        Args:
            machine: Machine instance with qmp_command method

        Returns:
            QMP command result
        """
        return machine.qmp_command("stop")

    @staticmethod
    def resume(machine) -> Dict[str, Any]:
        """
        Resume VM execution.

        Args:
            machine: Machine instance with qmp_command method

        Returns:
            QMP command result
        """
        return machine.qmp_command("cont")

    @staticmethod
    def device_del(machine, device_id: str) -> Dict[str, Any]:
        """
        Remove a hotplugged device.

        Args:
            machine: Machine instance with qmp_command method
            device_id: Device identifier to remove

        Returns:
            QMP command result
        """
        return machine.qmp_command("device_del", id=device_id)

    @staticmethod
    def system_powerdown(machine) -> Dict[str, Any]:
        """
        Send ACPI shutdown signal to VM.

        Args:
            machine: Machine instance with qmp_command method

        Returns:
            QMP command result
        """
        return machine.qmp_command("system_powerdown")

    @staticmethod
    def system_reset(machine) -> Dict[str, Any]:
        """
        Reset the VM.

        Args:
            machine: Machine instance with qmp_command method

        Returns:
            QMP command result
        """
        return machine.qmp_command("system_reset")



"""QMP Manager Protocol definitions.

Provides Protocol-based interface for QMP (QEMU Machine Protocol) operations.
Enables dependency inversion and eliminates circular dependencies.
"""

from __future__ import annotations

from typing import Any, Dict, List, Protocol, runtime_checkable


@runtime_checkable
class QMPManagerProtocol(Protocol):
    """Protocol for QMP (QEMU Machine Protocol) manager.

    Defines the interface for QMP operations used by other components.
    Supports both IPC and direct socket communication modes.

    This protocol is used for type hints to enable loose coupling between
    components without creating circular import dependencies.
    """

    def execute_qmp(
        self,
        vm_id: str,
        command: str,
        allow_dangerous: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Execute QMP command on VM.

        Args:
            vm_id: VM identifier (name or ID)
            command: QMP command name (e.g., "query-status", "system_powerdown")
            allow_dangerous: Allow dangerous commands (default: False)
            **kwargs: Command parameters

        Returns:
            QMP command result dictionary

        Raises:
            QMPError: If VM not found, not running, or command fails
        """
        ...

    def send_keys(
        self,
        vm_id: str,
        *keys: str,
        hold_time: int = 100,
    ) -> Dict[str, Any]:
        """Send key combination to VM via QMP.

        Args:
            vm_id: VM identifier (name or ID)
            *keys: Key names to press (e.g., 'ctrl', 'alt', 'f2')
            hold_time: How long to hold keys in milliseconds (default: 100)

        Returns:
            QMP command result dictionary

        Raises:
            QMPError: If VM not found, not running, or command fails
        """
        ...

    def type_text(
        self,
        vm_id: str,
        text: str,
        hold_time: int = 100,
    ) -> List[Dict[str, Any]]:
        """Type text string to VM via QMP.

        Args:
            vm_id: VM identifier (name or ID)
            text: Text to type
            hold_time: How long to hold each key in milliseconds (default: 100)

        Returns:
            List of QMP command results (one per character)

        Raises:
            QMPError: If VM not found, not running, or command fails
        """
        ...

    def take_screenshot(
        self,
        vm_id: str,
        filename: str,
    ) -> Dict[str, Any]:
        """Take screenshot of VM screen.

        Args:
            vm_id: VM identifier (name or ID)
            filename: Output filename for screenshot (e.g., "screenshot.ppm")

        Returns:
            QMP command result dictionary

        Raises:
            QMPError: If VM not found, not running, or command fails
        """
        ...

    def pause(self, vm_id: str) -> Dict[str, Any]:
        """Pause VM execution via QMP.

        Args:
            vm_id: VM identifier (name or ID)

        Returns:
            QMP command result dictionary

        Raises:
            QMPError: If VM not found, not running, or command fails
        """
        ...

    def resume(self, vm_id: str) -> Dict[str, Any]:
        """Resume VM execution via QMP.

        Args:
            vm_id: VM identifier (name or ID)

        Returns:
            QMP command result dictionary

        Raises:
            QMPError: If VM not found, not running, or command fails
        """
        ...

    def device_add(
        self,
        vm_id: str,
        driver: str,
        device_id: str,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Hot-plug device to VM via QMP.

        Args:
            vm_id: VM identifier (name or ID)
            driver: Device driver name (e.g., 'usb-storage', 'e1000', 'virtio-net-pci')
            device_id: Unique device identifier
            **kwargs: Additional device properties

        Returns:
            QMP command result dictionary

        Raises:
            QMPError: If VM not found, not running, or command fails
        """
        ...

    def device_del(
        self,
        vm_id: str,
        device_id: str,
    ) -> Dict[str, Any]:
        """Hot-unplug device from VM via QMP.

        Args:
            vm_id: VM identifier (name or ID)
            device_id: Device identifier to remove

        Returns:
            QMP command result dictionary

        Raises:
            QMPError: If VM not found, not running, or command fails
        """
        ...



"""
QMP Module

Contains QMP-related functionality for MAQET.
"""

from .commands import QMPCommands
from .keyboard import KeyboardEmulator
from .qmp_client import QMPClient, QMPClientError
from .qmp_socket_client import QMPSocketClient, QMPSocketError

__all__ = [
    "KeyboardEmulator",
    "QMPCommands",
    "QMPClient",
    "QMPClientError",
    "QMPSocketClient",
    "QMPSocketError",
]



"""
CLI Generator

Automatically generates argparse CLI commands from @api_method decorated methods.

This module implements the CLI interface generation component of MAQET's unified API system.
It takes method metadata and creates a complete command-line interface with proper argument
parsing, help text, and command routing.

The CLIGenerator converts Python method signatures into argparse subcommands, handling:
- Required and optional parameters
- Type conversion (str, int, bool, list)
- Default values and help text
- Command examples and documentation
- Error handling and validation

Path Handling:
- Functions accept Union[str, Path] for path parameters
- Internally uses pathlib.Path objects for all path operations
- Converts to str only when calling external APIs (subprocess, argparse)

Example:
    @api_method(cli_name="start", description="Start VM")
    def start(self, vm_id: str, detach: bool = False):
        pass

    Becomes CLI command:
    $ maqet start myvm --detach
"""

import argparse
import inspect
import logging
import sys
from pathlib import Path
from typing import Any, List, Optional, Union, get_args, get_origin

from ..api import APIMethodMetadata, APIRegistry
from ..config import RuntimeConfig
from ..constants import ExitCode
from ..exceptions import (
    VMNotFoundError,
    VMAlreadyExistsError,
    ConfigFileNotFoundError,
    ConfigValidationError,
    InvalidConfigurationError,
    SecurityError,
    SnapshotNotFoundError,
    WaitTimeout,
)
from ..logger import configure_file_logging, set_verbosity
from .base_generator import BaseGenerator

LOG = logging.getLogger(__name__)


class CLIGenerator(BaseGenerator):
    """
    Generates CLI commands from @api_method decorated methods.

    This generator creates an argparse-based CLI that automatically
    maps CLI arguments to method parameters and executes the appropriate
    method on the Maqet instance.

    Key Features:
    - Automatic subcommand generation from method metadata
    - Type-aware argument parsing (bool flags, optional args, etc.)
    - Built-in help generation with examples and descriptions
    - Global options support (--verbose, --config-dir, etc.)
    - Proper error handling and user feedback

    Usage:
        generator = CLIGenerator(maqet_instance, API_REGISTRY)
        result = generator.run(sys.argv[1:])

    The generator automatically creates CLI commands like:
    - maqet add config.yaml --name myvm
    - maqet start myvm --detach
    - maqet qmp myvm system_powerdown
    """

    def __init__(self, maqet_instance: Any, registry: APIRegistry):
        """
        Initialize CLI generator.

        Args:
            maqet_instance: Instance of Maqet class
            registry: API registry containing method metadata
        """
        super().__init__(maqet_instance, registry)
        self.parser: Optional[argparse.ArgumentParser] = None
        # Load runtime configuration for defaults
        self.runtime_config = RuntimeConfig()

    def generate(self) -> argparse.ArgumentParser:
        """
        Generate argparse CLI from registered methods.

        Returns:
            Configured ArgumentParser
        """
        # Create parent parser with global options for subparsers
        # This allows global options AFTER subcommands (e.g., maqet ls -v)
        self.global_parent = argparse.ArgumentParser(add_help=False)
        self._add_global_options_to_parent()

        # Create main parser with global options at top level
        # This allows global options BEFORE subcommands (e.g., maqet -v ls)
        self.parser = argparse.ArgumentParser(
            prog="maqet",
            description="MAQET - M4x0n's QEMU Tool for VM management",
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

        # Add global options directly to main parser (for before subcommand)
        self._add_global_options_to_main()

        # Add subcommands
        subparsers = self.parser.add_subparsers(
            dest="command", help="Available commands", metavar="COMMAND"
        )

        # Group methods by parent (None for top-level, string for nested)
        parent_groups = {}
        for category in self.registry.get_categories():
            methods = self.registry.get_by_category(category)
            for metadata in methods:
                if not metadata.hidden:
                    parent = metadata.parent
                    if parent not in parent_groups:
                        parent_groups[parent] = []
                    parent_groups[parent].append(metadata)

        # Add top-level commands (parent=None)
        if None in parent_groups:
            for metadata in parent_groups[None]:
                self._add_subcommand(subparsers, metadata)

        # Add parent commands with nested subcommands
        for parent, child_methods in parent_groups.items():
            if parent is not None:
                self._add_parent_command(subparsers, parent, child_methods)

        return self.parser

    def run(self, args: Optional[List[str]] = None) -> Any:
        """
        Parse arguments and execute the appropriate method.

        Args:
            args: Command line arguments (defaults to sys.argv[1:])

        Returns:
            Result of method execution
        """
        if self.parser is None:
            self.generate()

        parsed_args = self.parser.parse_args(args)

        # Configure logging based on verbosity flags
        self._configure_logging(parsed_args)

        if not hasattr(parsed_args, "command") or parsed_args.command is None:
            self.parser.print_help()
            sys.exit(ExitCode.INVALID_ARGS)

        # Determine which command to execute (handle nested subcommands)
        command_name = parsed_args.command

        # Check if this is a nested subcommand
        if hasattr(parsed_args, "subcommand") and parsed_args.subcommand:
            # This is a nested command (e.g., maqet qmp pause)
            metadata = self.registry.get_by_cli_name(parsed_args.subcommand)
            if not metadata:
                print(
                    f"Error: Unknown subcommand '{parsed_args.subcommand}'",
                    file=sys.stderr,
                )
                sys.exit(ExitCode.INVALID_ARGS)
        else:
            # This is a top-level command (e.g., maqet start)
            metadata = self.registry.get_by_cli_name(command_name)
            if not metadata:
                print(
                    f"Error: Unknown command '{command_name}'",
                    file=sys.stderr,
                )
                sys.exit(ExitCode.INVALID_ARGS)

        # Execute method
        try:
            result = self._execute_method(metadata, parsed_args)
            return result
        except Exception as e:
            cmd_display = (
                f"{command_name} {parsed_args.subcommand}"
                if hasattr(parsed_args, "subcommand") and parsed_args.subcommand
                else command_name
            )
            # Map exceptions to exit codes
            exit_code = self._get_exit_code_for_exception(e)
            print(
                f"Error executing {cmd_display}: {e}", file=sys.stderr
            )
            sys.exit(exit_code)

    def _add_global_options(self, parser: argparse.ArgumentParser) -> None:
        """
        Add global CLI options to any parser.

        These options are available both before and after subcommands.
        Defaults are loaded from maqet.conf if available.

        Args:
            parser: ArgumentParser to add options to
        """
        # Get defaults from runtime config
        verbosity_default = self.runtime_config.get_verbosity()
        config_dir_default = self.runtime_config.get_config_dir()
        data_dir_default = self.runtime_config.get_data_dir()
        runtime_dir_default = self.runtime_config.get_runtime_dir()
        log_file_default = self.runtime_config.get_log_file()

        # Build help messages that show config file source
        config_source = (
            f" (from {self.runtime_config.config_file_path})"
            if self.runtime_config.config_file_path
            else ""
        )

        parser.add_argument(
            "-v",
            "--verbose",
            action="count",
            default=verbosity_default,
            help=f"Increase verbosity: -v=warnings+errors, -vv=info, -vvv=debug "
            f"(default: {verbosity_default}=errors only{config_source})",
        )
        # Directory flags are pre-parsed in __main__.py before Maqet creation
        # to allow proper CLI > config file > XDG precedence order
        parser.add_argument(
            "--maqet-config-dir",
            default=config_dir_default,
            help="Override maqet's config directory path"
            + (f" (default: {config_dir_default})" if config_dir_default else ""),
        )
        parser.add_argument(
            "--maqet-data-dir",
            default=data_dir_default,
            help="Override maqet's data directory path"
            + (f" (default: {data_dir_default})" if data_dir_default else ""),
        )
        parser.add_argument(
            "--maqet-runtime-dir",
            default=runtime_dir_default,
            help="Override maqet's runtime directory path"
            + (f" (default: {runtime_dir_default})" if runtime_dir_default else ""),
        )
        parser.add_argument(
            "--log-file",
            default=log_file_default,
            help="Enable file logging to specified path"
            + (f" (default: {log_file_default})" if log_file_default else ""),
        )

    def _add_global_options_to_parent(self) -> None:
        """
        Add global CLI options to parent parser (inherited by all subcommands).
        """
        self._add_global_options(self.global_parent)

    def _add_global_options_to_main(self) -> None:
        """
        Add global CLI options to main parser (before subcommand).
        """
        self._add_global_options(self.parser)

    def _configure_logging(self, args: argparse.Namespace) -> None:
        """
        Configure logging based on CLI arguments.

        Verbosity mapping:
        - 0 (no -v):    ERROR level (default, errors only)
        - 1 (-v):       WARNING level (shows warnings + errors)
        - 2 (-vv):      INFO level
        - 3+ (-vvv+):   DEBUG level

        Args:
            args: Parsed command line arguments
        """
        verbosity = getattr(args, "verbose", 0)

        # Direct mapping: verbosity count = logger level
        set_verbosity(verbosity)

        # File logging setup (unchanged)
        log_file = getattr(args, "log_file", None)
        if log_file:
            from pathlib import Path

            configure_file_logging(Path(log_file))

    def _add_subcommand(
        self,
        subparsers: argparse._SubParsersAction,
        metadata: APIMethodMetadata,
    ) -> None:
        """
        Add a subcommand for a method.

        Args:
            subparsers: Argparse subparsers object
            metadata: Method metadata
        """
        # Create subparser with global parent to inherit global options
        sub = subparsers.add_parser(
            metadata.cli_name,
            help=metadata.description,
            description=metadata.description,
            formatter_class=argparse.RawDescriptionHelpFormatter,
            parents=[self.global_parent],
        )

        # Add aliases if specified
        for alias in metadata.aliases:
            subparsers._name_parser_map[alias] = sub

        # Add method parameters as arguments (excluding **kwargs)
        for param_name, param in metadata.parameters.items():
            if param.kind != inspect.Parameter.VAR_KEYWORD:  # Skip **kwargs
                self._add_parameter_argument(sub, metadata, param_name, param)

        # Special handling for apply command: add common config parameters
        if metadata.cli_name == "apply":
            self._add_apply_config_parameters(sub)

        # Add examples to help if available
        if metadata.examples:
            examples_text = "\nExamples:\n" + "\n".join(
                f"  {example}" for example in metadata.examples
            )
            sub.epilog = examples_text

    def _add_parent_command(
        self,
        subparsers: argparse._SubParsersAction,
        parent_name: str,
        child_methods: List[APIMethodMetadata],
    ) -> None:
        """
        Add a parent command with nested subcommands.

        Args:
            subparsers: Argparse subparsers object
            parent_name: Name of the parent command (e.g., 'qmp')
            child_methods: List of child method metadata
        """
        # Create parent subparser with global parent to inherit global options
        parent_parser = subparsers.add_parser(
            parent_name,
            help=f"{parent_name.upper()} subcommands",
            description=f"{parent_name.upper()} subcommands",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            parents=[self.global_parent],
        )

        # Add nested subparsers for child commands
        child_subparsers = parent_parser.add_subparsers(
            dest="subcommand",
            help="Available subcommands",
            metavar="SUBCOMMAND",
        )

        # Add each child method as a nested subcommand
        for metadata in child_methods:
            self._add_subcommand(child_subparsers, metadata)

    def _add_parameter_argument(
        self,
        parser: argparse.ArgumentParser,
        metadata: APIMethodMetadata,
        param_name: str,
        param: inspect.Parameter,
    ) -> None:
        """
        Add an argument for a method parameter.

        Args:
            parser: Argument parser
            metadata: Method metadata for context
            param_name: Parameter name
            param: Parameter metadata
        """
        # Determine argument properties
        is_required = param.default == inspect.Parameter.empty
        arg_name = param_name.replace("_", "-")

        # Check for Union[str, List[str]] type for multiple files support
        is_multiple_files = self._is_multiple_files_param(param)

        # Special handling for vm_id parameters - make them positional for
        # better UX
        is_vm_id_param = param_name == "vm_id"

        # For rm command, vm_id should be optional positional since it has
        # --all alternative
        is_rm_command = metadata.cli_name == "rm"

        # Special handling for VAR_POSITIONAL parameters (*args)
        is_var_positional = param.kind == inspect.Parameter.VAR_POSITIONAL

        if param.annotation == bool:
            # Boolean flags
            if is_required:
                # Required boolean (rare case)
                parser.add_argument(
                    f"--{arg_name}",
                    action="store_true",
                    required=True,
                    help=f"{param_name} (required boolean flag)",
                )
            else:
                # Optional boolean flag
                default_value = (
                    param.default
                    if param.default != inspect.Parameter.empty
                    else False
                )
                parser.add_argument(
                    f"--{arg_name}",
                    action=(
                        "store_true" if not default_value else "store_false"
                    ),
                    default=default_value,
                    help=f"{param_name} (default: {default_value})",
                )
        elif is_multiple_files:
            # Multiple files parameter (Union[str, List[str]])
            if is_required:
                parser.add_argument(
                    param_name,
                    nargs="+",
                    help=f"{param_name} (one or more files)",
                )
            else:
                parser.add_argument(
                    f"--{arg_name}",
                    nargs="*",
                    default=param.default,
                    help=f"{param_name} (one or more files, default: {
                        param.default})",
                )
        elif is_vm_id_param:
            # Special handling for vm_id parameters - make them positional for
            # better UX
            if is_rm_command:
                # For rm command, vm_id is optional positional since --all is
                # alternative
                parser.add_argument(
                    param_name,
                    nargs="?",
                    default=param.default,
                    help=f"{
                        param_name} (VM name or ID, optional when using --all)",
                )
            elif is_required:
                # Regular required vm_id (for start, stop, status, etc.)
                parser.add_argument(
                    param_name, help=f"{param_name} (VM name or ID, required)"
                )
            else:
                # Optional vm_id but still positional
                parser.add_argument(
                    param_name,
                    nargs="?",
                    default=param.default,
                    help=f"{param_name} (VM name or ID, optional)",
                )
        elif is_var_positional:
            # VAR_POSITIONAL parameters (*args) - use nargs='*' for zero or
            # more
            parser.add_argument(
                param_name,
                nargs="*",
                help=f"{param_name} (zero or more values)",
            )
        elif is_required:
            # Required positional argument
            # Extract type from annotation for proper parsing
            arg_type = self._get_arg_type(param)
            if arg_type:
                parser.add_argument(
                    param_name, type=arg_type, help=f"{param_name} (required)"
                )
            else:
                parser.add_argument(param_name, help=f"{param_name} (required)")
        else:
            # Optional argument with default
            # Extract type from annotation for proper parsing
            arg_type = self._get_arg_type(param)
            if arg_type:
                parser.add_argument(
                    f"--{arg_name}",
                    type=arg_type,
                    default=param.default,
                    help=f"{param_name} (default: {param.default})",
                )
            else:
                parser.add_argument(
                    f"--{arg_name}",
                    default=param.default,
                    help=f"{param_name} (default: {param.default})",
                )

    def _get_arg_type(self, param: inspect.Parameter) -> Optional[type]:
        """
        Extract concrete type from parameter annotation for argparse.

        Handles Optional[T] by extracting T, and returns basic types
        that argparse can handle (int, float, str).

        Args:
            param: Parameter to extract type from

        Returns:
            Type for argparse, or None if no type annotation or complex type
        """
        if param.annotation == inspect.Parameter.empty:
            return None

        # Handle Optional[T] (which is Union[T, None])
        origin = get_origin(param.annotation)
        if origin is Union:
            args = get_args(param.annotation)
            # Filter out NoneType from Optional[T]
            non_none_types = [arg for arg in args if arg is not type(None)]
            if len(non_none_types) == 1:
                # This is Optional[T], use T
                annotation = non_none_types[0]
            else:
                # Complex union, can't determine single type
                return None
        else:
            annotation = param.annotation

        # Return basic types that argparse supports
        if annotation in (int, float, str):
            return annotation

        return None

    def _is_multiple_files_param(self, param: inspect.Parameter) -> bool:
        """
        Check if parameter accepts multiple files (Union[str, List[str]]).

        Args:
            param: Parameter to check

        Returns:
            True if parameter accepts multiple files
        """
        if param.annotation == inspect.Parameter.empty:
            return False

        # Check for Union[str, List[str]] or similar patterns
        origin = get_origin(param.annotation)
        if origin is Union:
            args = get_args(param.annotation)
            # Check for Union[str, List[str]] pattern
            has_str = str in args
            has_list_str = any(
                get_origin(arg) is list and get_args(arg) == (str,)
                for arg in args
            )
            return has_str and has_list_str

        return False

    def _execute_method(
        self, metadata: APIMethodMetadata, args: argparse.Namespace
    ) -> Any:
        """
        Execute a method with parsed arguments.

        Args:
            metadata: Method metadata
            args: Parsed command line arguments

        Returns:
            Method execution result
        """
        # Extract method parameters from parsed args
        method_kwargs = {}
        method_args = []

        # Check if method has VAR_POSITIONAL (*args) parameter
        has_var_positional = any(
            param.kind == inspect.Parameter.VAR_POSITIONAL
            for param in metadata.parameters.values()
        )

        # Collect positional parameters that come before VAR_POSITIONAL
        positional_params = []
        for param_name, param in metadata.parameters.items():
            if param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                positional_params.append((param_name, param))
            elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                break  # Stop when we hit VAR_POSITIONAL

        for param_name, param in metadata.parameters.items():
            # Skip **kwargs parameters
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                continue

            # Convert parameter names (dashes to underscores)
            arg_name = param_name.replace("_", "-")

            # Get value from args
            if hasattr(args, param_name):
                value = getattr(args, param_name)
            elif hasattr(args, arg_name):
                value = getattr(args, arg_name.replace("-", "_"))
            else:
                continue

            if value is not None:
                # Handle VAR_POSITIONAL parameters (*args) specially
                if param.kind == inspect.Parameter.VAR_POSITIONAL:
                    # For *args parameters, extend the args list
                    if isinstance(value, list):
                        method_args.extend(value)
                    else:
                        method_args.append(value)
                elif (
                    param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
                    and has_var_positional
                ):
                    # When method has VAR_POSITIONAL, regular positional params
                    # must go in method_args to avoid "multiple values" error
                    method_args.insert(0, value)
                else:
                    # Everything else goes in kwargs
                    method_kwargs[param_name] = value

        # Special handling for apply command: collect config parameters as
        # kwargs
        if metadata.cli_name == "apply":
            config_params = {}
            for param_name in ["memory", "cpu", "binary", "enable_kvm"]:
                arg_name = param_name.replace("_", "-")
                if hasattr(args, arg_name.replace("-", "_")):
                    value = getattr(args, arg_name.replace("-", "_"))
                    if value is not None:
                        config_params[param_name] = value

            # Add config parameters to method kwargs
            method_kwargs.update(config_params)

        # Resolve config file paths to absolute paths for consistent handling
        if "config" in method_kwargs and method_kwargs["config"] is not None:

            config = method_kwargs["config"]
            if isinstance(config, str):
                # Single config file - resolve to absolute path
                method_kwargs["config"] = str(Path(config).resolve())
            elif isinstance(config, list):
                # Multiple config files - resolve each to absolute path
                method_kwargs["config"] = [str(Path(c).resolve()) for c in config]

        # Execute method directly
        method = getattr(self.maqet_instance, metadata.name)

        # Execute method with both positional and keyword arguments
        return method(*method_args, **method_kwargs)

    def _add_apply_config_parameters(
        self, parser: argparse.ArgumentParser
    ) -> None:
        """
        Add common configuration parameters for the apply command.

        Args:
            parser: Argument parser for the apply subcommand
        """
        # Memory configuration
        parser.add_argument(
            "--memory", type=str, help="VM memory size (e.g., 2G, 4096M)"
        )

        # CPU configuration
        parser.add_argument("--cpu", type=int, help="Number of CPU cores")

        # Binary path
        parser.add_argument("--binary", type=str, help="Path to QEMU binary")

        # KVM enablement
        parser.add_argument(
            "--enable-kvm", action="store_true", help="Enable KVM acceleration"
        )

    def _get_exit_code_for_exception(self, exception: Exception) -> int:
        """
        Map exception types to appropriate exit codes.

        This method implements the exit code convention for consistent
        error reporting across all maqet commands.

        Args:
            exception: Exception that was raised

        Returns:
            Appropriate exit code based on exception type

        Exit Code Mapping:
            - ExitCode.INVALID_ARGS (3): Invalid arguments or preconditions not met
              * VMNotFoundError
              * VMAlreadyExistsError
              * SnapshotNotFoundError
              * ConfigFileNotFoundError
              * ConfigValidationError
              * InvalidConfigurationError
            - ExitCode.PERMISSION_DENIED (4): Permission errors
              * SecurityError
              * PermissionError
              * OSError with permission-related errno
            - ExitCode.TIMEOUT (2): Timeout conditions
              * WaitTimeout
            - ExitCode.FAILURE (1): All other errors
              * DatabaseLockError
              * MaqetError
              * Unexpected exceptions
        """
        # INVALID_ARGS (3): Invalid arguments or preconditions not met
        if isinstance(exception, (
            VMNotFoundError,
            VMAlreadyExistsError,
            SnapshotNotFoundError,
            ConfigFileNotFoundError,
            ConfigValidationError,
            InvalidConfigurationError,
        )):
            return ExitCode.INVALID_ARGS

        # PERMISSION_DENIED (4): Permission-related errors
        if isinstance(exception, (SecurityError, PermissionError)):
            return ExitCode.PERMISSION_DENIED

        # Check for OSError with permission-related errno
        if isinstance(exception, OSError):
            import errno
            if exception.errno in (errno.EACCES, errno.EPERM):
                return ExitCode.PERMISSION_DENIED

        # TIMEOUT (2): Timeout conditions
        if isinstance(exception, WaitTimeout):
            return ExitCode.TIMEOUT

        # FAILURE (1): Database locks and other failures
        # Default for all other MaqetError and unexpected exceptions
        return ExitCode.FAILURE



"""
Process validation and monitoring utilities.

This module provides utilities for checking process existence and
monitoring process lifecycle. Centralizes PID validation patterns
to ensure consistent error handling across the codebase.

Key utilities:
- validate_pid_exists(): Check if a process exists by PID
- wait_for_process_exit(): Wait for a process to terminate

Usage:
    from maqet.utils.process_validation import validate_pid_exists

    # Raise exception if QEMU process doesn't exist
    validate_pid_exists(qemu_pid, "QEMU process", raise_on_missing=True)
"""

import os
import time
import logging

from ..exceptions import ProcessNotFoundError

LOG = logging.getLogger(__name__)


def validate_pid_exists(
    pid: int,
    process_name: str = "process",
    raise_on_missing: bool = False,
    log_permission_warning: bool = True
) -> bool:
    """
    Check if a process exists by PID using os.kill signal 0.

    Signal 0 is a special case that doesn't actually send a signal but
    performs error checking on whether the process exists.

    Args:
        pid: Process ID to check
        process_name: Descriptive name for error messages (e.g., "QEMU process")
        raise_on_missing: If True, raise ProcessNotFoundError instead of returning False
        log_permission_warning: If True, log warning when PermissionError occurs

    Returns:
        True if process exists (or permission denied), False if process not found

    Raises:
        ProcessNotFoundError: If raise_on_missing=True and process doesn't exist
        ValueError: If PID is invalid (<= 0)

    Note:
        PermissionError is treated as "exists" because we can't verify but
        the error indicates the process is running (owned by different user).
        This should not happen in normal maqet usage since we check our own processes.

    Examples:
        >>> validate_pid_exists(12345, "QEMU", raise_on_missing=True)
        True

        >>> validate_pid_exists(99999)  # Non-existent PID
        False

        >>> validate_pid_exists(1, raise_on_missing=True)  # init process, permission denied
        True  # Exists but can't verify
    """
    if not pid or pid <= 0:
        if raise_on_missing:
            raise ValueError(f"Invalid PID: {pid} (must be > 0)")
        return False

    try:
        os.kill(pid, 0)  # Signal 0 = check existence without killing
        LOG.debug(f"{process_name} (PID {pid}) verified alive")
        return True

    except ProcessLookupError:
        # Process definitively does not exist
        LOG.debug(f"{process_name} (PID {pid}) not found")

        if raise_on_missing:
            raise ProcessNotFoundError(
                f"{process_name} (PID {pid}) does not exist"
            )
        return False

    except PermissionError:
        # Process exists but we don't have permission to signal it
        # This should NOT happen for our own QEMU processes
        if log_permission_warning:
            LOG.warning(
                f"Permission denied checking {process_name} (PID {pid}). "
                f"Process exists but owned by different user (unexpected)."
            )
        # Treat as "exists" since error confirms process is running
        return True


def wait_for_process_exit(
    pid: int,
    timeout: float,
    check_interval: float = 0.1,
    process_name: str = "process"
) -> bool:
    """
    Wait for a process to exit, checking periodically.

    Args:
        pid: Process ID to monitor
        timeout: Maximum seconds to wait
        check_interval: Seconds between existence checks
        process_name: Descriptive name for logging

    Returns:
        True if process exited, False if timeout reached

    Examples:
        >>> wait_for_process_exit(12345, timeout=5.0, process_name="QEMU")
        True  # Process exited within 5 seconds
    """
    elapsed = 0.0
    while elapsed < timeout:
        if not validate_pid_exists(pid, process_name=process_name):
            LOG.debug(f"{process_name} (PID {pid}) exited after {elapsed:.2f}s")
            return True

        time.sleep(check_interval)
        elapsed += check_interval

    LOG.warning(f"{process_name} (PID {pid}) still running after {timeout}s timeout")
    return False



"""
Unix Socket IPC Server

Simple Unix domain socket server for IPC between CLI and VM runner.
Uses JSON-RPC style protocol for request/response communication.

Architecture:
- Each VM runner process has its own Unix socket
- Non-blocking async I/O using asyncio
- Simple request/response pattern
- Socket cleanup on server stop
- Challenge-response authentication for security
"""

import asyncio
import hashlib
import hmac
import json
import os
import secrets
import stat
from pathlib import Path
from typing import Any, Callable, Dict, Optional

from ..constants import Timeouts
from ..logger import LOG


class UnixSocketIPCServerError(Exception):
    """Unix socket IPC server errors."""


class UnixSocketIPCServer:
    """
    Unix domain socket server for IPC between CLI and VM runner.

    Protocol: JSON-RPC style
    - Client sends: {"method": "qmp", "args": [...], "kwargs": {...}}
    - Server responds: {"status": "success", "result": ...} or
                       {"status": "error", "error": "..."}

    Socket lifecycle:
    1. Server starts, binds to socket path
    2. Accepts connections from CLI clients
    3. Reads JSON request, calls handler
    4. Writes JSON response
    5. Closes connection
    """

    def __init__(
        self,
        socket_path: Path,
        handler: Callable[[Dict[str, Any]], Dict[str, Any]],
        auth_secret: Optional[str] = None
    ):
        """
        Initialize Unix socket server with authentication support.

        Args:
            socket_path: Path to Unix socket file
            handler: Async function to handle requests
                     Takes request dict, returns response dict
            auth_secret: Optional authentication secret for challenge-response
        """
        self.socket_path = Path(socket_path)
        self.handler = handler
        self.auth_secret = auth_secret
        self.server: Optional[asyncio.Server] = None
        self._running = False
        self._loop: Optional[asyncio.AbstractEventLoop] = None

        LOG.debug(f"UnixSocketIPCServer initialized for {socket_path}")

    async def start(self) -> None:
        """
        Start listening on Unix socket with secure permissions.

        Process:
        1. Remove existing socket if present (stale socket handling)
        2. Set restrictive umask before socket creation
        3. Create Unix socket server
        4. Verify socket permissions are secure (0600)
        5. Start accepting connections
        6. Keep server running until stop() called

        Security:
        - Socket created with 0600 permissions (user-only access)
        - Prevents local privilege escalation (CVSS 7.8)
        - Original umask restored after socket creation

        Raises:
            UnixSocketIPCServerError: If socket already in use or bind fails
        """
        # Store event loop for cross-thread communication
        self._loop = asyncio.get_running_loop()

        # Remove existing socket if present
        if self.socket_path.exists():
            # Try to connect to check if someone is using it
            try:
                reader, writer = await asyncio.wait_for(
                    asyncio.open_unix_connection(str(self.socket_path)),
                    timeout=Timeouts.IPC_HEALTH_CHECK,
                )
                writer.close()
                await writer.wait_closed()
                # Someone is using it
                raise UnixSocketIPCServerError(
                    f"Socket already in use: {self.socket_path}"
                )
            except (ConnectionRefusedError, FileNotFoundError, asyncio.TimeoutError):
                # Stale socket, remove it
                LOG.debug(f"Removing stale socket {self.socket_path}")
                self.socket_path.unlink()

        # Set restrictive umask before socket creation
        # This ensures only the owner can access the socket
        old_umask = os.umask(0o077)

        try:
            # Create Unix socket server
            self.server = await asyncio.start_unix_server(
                self._handle_client, path=str(self.socket_path)
            )
            self._running = True

            # Verify socket has correct permissions (0600 for socket files)
            socket_stat = self.socket_path.stat()
            expected_mode = stat.S_IRUSR | stat.S_IWUSR  # 0o600
            actual_mode = stat.S_IMODE(socket_stat.st_mode)

            if actual_mode != expected_mode:
                LOG.warning(
                    f"Socket permissions {oct(actual_mode)} differ from "
                    f"expected {oct(expected_mode)}. Attempting to fix."
                )
                os.chmod(self.socket_path, expected_mode)

            LOG.info(
                f"IPC server listening on {self.socket_path} (mode: 0600)"
            )

            # Keep server running
            async with self.server:
                await self.server.serve_forever()

        except asyncio.CancelledError:
            # Server was stopped via stop() - this is normal
            LOG.debug("IPC server cancelled (normal shutdown)")
        except Exception as e:
            raise UnixSocketIPCServerError(f"Failed to start IPC server: {e}")
        finally:
            # Restore original umask
            os.umask(old_umask)

    async def _handle_client(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        """
        Handle single client connection with authentication.

        Process:
        1. (If auth enabled) Perform challenge-response authentication
        2. Read JSON request from client
        3. Parse and validate request
        4. Call handler function
        5. Write JSON response to client
        6. Close connection

        Args:
            reader: Async stream reader
            writer: Async stream writer
        """
        try:
            # Step 1: Authentication (if enabled)
            if self.auth_secret:
                if not await self._authenticate_client(reader, writer):
                    LOG.warning("Client authentication failed")
                    return

            # Step 2: Read request with newline framing (consistent with auth)
            try:
                data = await asyncio.wait_for(
                    reader.readuntil(b'\n'), timeout=Timeouts.IPC_COMMAND
                )
            except asyncio.LimitOverrunError:
                LOG.warning("Request message too long or missing newline delimiter")
                response = {"status": "error", "error": "Request too large or malformed"}
                writer.write((json.dumps(response) + "\n").encode("utf-8"))
                await writer.drain()
                return
            except asyncio.TimeoutError:
                LOG.warning("Request read timeout")
                return

            if not data:
                return

            # Step 3: Parse JSON request
            try:
                request = json.loads(data.decode("utf-8").strip())
            except json.JSONDecodeError as e:
                response = {"status": "error", "error": f"Invalid JSON: {e}"}
                writer.write((json.dumps(response) + "\n").encode("utf-8"))
                await writer.drain()
                return

            LOG.debug(f"IPC request: {request.get('method', 'unknown')}")

            # Step 4: Call handler
            try:
                response = await self.handler(request)
            except Exception as e:
                LOG.error(f"Handler error: {e}")
                response = {"status": "error", "error": str(e)}

            # Step 5: Write response with newline framing (consistent protocol)
            response_data = (json.dumps(response) + "\n").encode("utf-8")
            writer.write(response_data)
            await writer.drain()

            LOG.debug(f"IPC response: {response.get('status', 'unknown')}")

        except Exception as e:
            LOG.error(f"Error handling client: {e}")

        finally:
            # Close connection
            try:
                writer.close()
                await writer.wait_closed()
            except Exception as e:
                LOG.debug(f"Error closing connection: {e}")

    async def _authenticate_client(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> bool:
        """
        Perform challenge-response authentication with client.

        Protocol:
        1. Server sends challenge: {"type": "challenge", "value": "<random-hex>"}
        2. Client computes HMAC-SHA256(auth_secret, challenge)
        3. Client sends: {"auth": "<hmac-hex>"}
        4. Server verifies HMAC matches expected value

        Args:
            reader: Async stream reader
            writer: Async stream writer

        Returns:
            True if authentication successful, False otherwise
        """
        try:
            # Generate 128-bit random challenge
            challenge = secrets.token_hex(16)

            # Send challenge to client with newline framing (consistent protocol)
            challenge_msg = json.dumps({"type": "challenge", "value": challenge}) + "\n"
            writer.write(challenge_msg.encode("utf-8"))
            await writer.drain()

            LOG.debug("Sent authentication challenge")

            # Receive response from client
            # Use readuntil() with newline delimiter for proper message framing
            # This ensures we don't consume data beyond the auth response
            try:
                response_data = await asyncio.wait_for(
                    reader.readuntil(b'\n'), timeout=Timeouts.IPC_AUTH
                )
            except asyncio.LimitOverrunError:
                # Message too long or no delimiter found
                LOG.warning("Authentication response too long or malformed")
                return False

            if not response_data:
                LOG.warning("Empty authentication response")
                return False

            # Parse response
            try:
                response = json.loads(response_data.decode("utf-8").strip())
            except json.JSONDecodeError:
                LOG.warning("Invalid JSON in authentication response")
                return False

            # Verify HMAC response
            expected_hmac = hmac.new(
                self.auth_secret.encode("utf-8"),
                challenge.encode("utf-8"),
                hashlib.sha256
            ).hexdigest()

            client_hmac = response.get("auth", "")

            # Use constant-time comparison to prevent timing attacks
            if not hmac.compare_digest(expected_hmac, client_hmac):
                LOG.warning("Authentication failed: HMAC mismatch")
                error_msg = json.dumps({"status": "error", "error": "Authentication failed"})
                writer.write(error_msg.encode("utf-8"))
                await writer.drain()
                return False

            LOG.debug("Client authenticated successfully")
            return True

        except asyncio.TimeoutError:
            LOG.warning("Authentication timeout")
            return False
        except Exception as e:
            LOG.error(f"Authentication error: {e}")
            return False

    async def stop(self) -> None:
        """
        Stop server and cleanup socket (async version).

        Process:
        1. Close server (stop accepting connections)
        2. Remove socket file

        Note: This should be called from within the same event loop as start().
        For cross-thread stopping, use stop_sync() instead.
        """
        LOG.debug("Stopping IPC server")
        self._running = False

        # Close server
        if self.server:
            self.server.close()
            await self.server.wait_closed()

        # Remove socket file
        if self.socket_path.exists():
            try:
                self.socket_path.unlink()
                LOG.debug(f"Removed socket {self.socket_path}")
            except Exception as e:
                LOG.warning(f"Failed to remove socket: {e}")

    def stop_sync(self) -> None:
        """
        Stop server from another thread (synchronous).

        This method is thread-safe and can be called from the main thread
        to stop the IPC server running in a background thread.

        Process:
        1. Mark server as stopped
        2. Close server socket (cancels serve_forever)
        3. Remove socket file from filesystem
        """
        LOG.debug("Stopping IPC server (sync)")
        self._running = False

        # Close server socket from any thread
        # This will cause serve_forever() to raise CancelledError
        if self.server:
            # Call server.close() in a thread-safe way
            if self._loop and self._loop.is_running():
                # Schedule close in the IPC server's event loop
                self._loop.call_soon_threadsafe(self.server.close)
            else:
                # Event loop not running, close directly
                self.server.close()

        # Remove socket file (filesystem operation, thread-safe)
        if self.socket_path.exists():
            try:
                self.socket_path.unlink()
                LOG.debug(f"Removed socket {self.socket_path}")
            except Exception as e:
                LOG.warning(f"Failed to remove socket: {e}")

    @property
    def is_running(self) -> bool:
        """Check if server is running."""
        return self._running



"""
Process management utilities.

Provides PID verification, process token management, and safe process operations.

Process Verification Guide:
    1. Simple existence check: validate_pid_exists() from process_validation.py
    2. Token-based verification: ProcessVerifier.verify_or_raise() from this module
    3. /proc inspection: ProcFS class from this module

Use validate_pid_exists() for basic "is process alive" checks.
Use ProcessVerifier for safety-critical operations requiring PID reuse protection.
Use ProcFS for low-level /proc filesystem access.
"""

from typing import Dict, Optional, List, Type
from dataclasses import dataclass

try:
    import psutil  # noqa: F401
    PSUTIL_AVAILABLE = True
except ImportError:
    PSUTIL_AVAILABLE = False

from maqet.logger import LOG


class ProcFS:
    """Linux /proc filesystem utilities.

    Provides low-level access to process information via /proc filesystem.
    Linux-only. All methods return None on error (process not found, permission denied).

    Example:
        # Get process command line
        cmdline = ProcFS.get_cmdline(1234)

        # Check if process is zombie
        if ProcFS.is_zombie(1234):
            print("Process is zombie")

        # Get process state
        state = ProcFS.get_state(1234)  # "R", "S", "D", "Z", etc.
    """

    @staticmethod
    def get_cmdline(pid: int) -> Optional[List[str]]:
        """Get process command line from /proc/{pid}/cmdline.

        Args:
            pid: Process ID

        Returns:
            List of command line arguments, or None if process not found/permission denied
        """
        try:
            with open(f"/proc/{pid}/cmdline", "rb") as f:
                data = f.read()
                if not data:
                    return None
                return data.decode().rstrip('\0').split('\0')
        except (OSError, IOError):
            return None

    @staticmethod
    def get_status(pid: int) -> Optional[Dict[str, str]]:
        """Get process status from /proc/{pid}/status.

        Args:
            pid: Process ID

        Returns:
            Dict of status fields (Name, State, PPid, etc.), or None if not found
        """
        try:
            with open(f"/proc/{pid}/status") as f:
                result = {}
                for line in f:
                    if ':' in line:
                        key, value = line.split(':', 1)
                        result[key.strip()] = value.strip()
                return result
        except (OSError, IOError):
            return None

    @staticmethod
    def get_state(pid: int) -> Optional[str]:
        """Get process state character from /proc/{pid}/stat.

        Process states:
        - R: Running
        - S: Sleeping (interruptible)
        - D: Disk sleep (uninterruptible)
        - Z: Zombie
        - T: Stopped
        - t: Tracing stop
        - X: Dead
        - I: Idle

        Args:
            pid: Process ID

        Returns:
            Single character state, or None if process not found
        """
        try:
            with open(f"/proc/{pid}/stat", "r") as f:
                stat = f.read()
                # State is after closing paren: "pid (comm) S ..."
                state_start = stat.rfind(")") + 2
                return stat[state_start] if state_start < len(stat) else None
        except (FileNotFoundError, IOError):
            return None

    @staticmethod
    def is_zombie(pid: int) -> bool:
        """Check if process is a zombie (state Z).

        Args:
            pid: Process ID

        Returns:
            True if process is zombie, False otherwise (including not found)
        """
        return ProcFS.get_state(pid) == "Z"

    @staticmethod
    def get_ppid(pid: int) -> Optional[int]:
        """Get parent process ID.

        Args:
            pid: Process ID

        Returns:
            Parent PID, or None if process not found
        """
        status = ProcFS.get_status(pid)
        if not status:
            return None
        try:
            return int(status.get("PPid", ""))
        except ValueError:
            return None


@dataclass
class ProcessVerification:
    """Result of process verification."""
    pid: int
    is_verified: bool
    process_name: str
    cmdline: str
    error: Optional[str] = None


class ProcessVerifier:
    """
    Centralized process verification with error handling.

    Wraps verify_process() calls with consistent error handling patterns.
    Converts low-level ValueError exceptions into domain-specific exceptions.
    """

    @staticmethod
    def verify_or_raise(
        pid: int,
        expected_names: List[str],
        expected_tokens: List[str],
        error_type: Type[Exception],
        process_description: str = "process"
    ) -> bool:
        """
        Verify process identity and raise domain-specific exception on failure.

        Provides centralized error handling for process verification:
        1. Calls verify_process() with provided parameters
        2. Handles "does not exist" case (returns False)
        3. Converts verification failures to domain-specific exceptions

        Args:
            pid: Process ID to verify
            expected_names: List of acceptable process names (e.g., ["python", "python3"])
            expected_tokens: Tokens that must appear in cmdline (e.g., ["maqet", "vm_runner"])
            error_type: Exception class to raise on verification failure
            process_description: Human-readable process description for error messages

        Returns:
            True if process exists and verified, False if process doesn't exist

        Raises:
            error_type: If process exists but verification fails (wrong process)

        Example:
            # QEMU verification
            ProcessVerifier.verify_or_raise(
                qemu_pid,
                expected_names=["qemu", "qemu-system-x86_64"],
                expected_tokens=[vm_id, vm_name],
                error_type=VMLifecycleError,
                process_description=f"QEMU for VM '{vm_id}'"
            )

            # Runner verification
            ProcessVerifier.verify_or_raise(
                runner_pid,
                expected_names=["python", "python3"],
                expected_tokens=["maqet", "vm_runner"],
                error_type=RunnerSpawnError,
                process_description=f"VM runner (PID {runner_pid})"
            )
        """
        try:
            verify_process(pid, expected_names, expected_tokens)
            LOG.debug(f"Verified PID {pid} is {process_description}. Safe to operate on.")
            return True

        except ValueError as e:
            error_msg = str(e)

            # Process doesn't exist - return False (not an error)
            if "does not exist" in error_msg:
                LOG.debug(f"{process_description} (PID {pid}) does not exist")
                return False

            # Process exists but wrong type - raise domain exception
            LOG.error(f"PID verification failed for {process_description}: {error_msg}")
            raise error_type(
                f"Stale PID {pid} does not match expected {process_description}. "
                f"Manual cleanup required. Details: {error_msg}"
            )


def verify_process(
    pid: int,
    expected_names: List[str],
    expected_cmdline_tokens: List[str],
    warn_recent: float = 1.0
) -> ProcessVerification:
    """
    Verify process identity with PID reuse protection.

    Performs multi-layer verification:
    1. Process name matches expected names
    2. Command line contains expected tokens
    3. (Optional) Warns if process very recently created

    Args:
        pid: Process ID to verify
        expected_names: List of acceptable process names (e.g., ["python", "python3"])
        expected_cmdline_tokens: Tokens that must appear in cmdline (e.g., ["maqet", "vm_runner"])
        warn_recent: Warn if process created within this many seconds (default 1.0)

    Returns:
        ProcessVerification with verification result

    Raises:
        ValueError: If process exists but doesn't match verification criteria

    Example:
        # Verify QEMU process
        result = verify_process(
            qemu_pid,
            expected_names=["qemu-system-x86_64", "qemu"],
            expected_cmdline_tokens=[vm_id, vm_name]
        )

        # Verify runner process
        result = verify_process(
            runner_pid,
            expected_names=["python", "python3"],
            expected_cmdline_tokens=["maqet", "vm_runner"]
        )
    """
    if PSUTIL_AVAILABLE:
        try:
            import psutil
            process = psutil.Process(pid)

            # Check 1: Process name matches
            process_name = process.name().lower()
            name_match = any(name.lower() in process_name for name in expected_names)

            if not name_match:
                error = (
                    f"PID {pid} process name '{process.name()}' does not match "
                    f"expected names {expected_names}. Possible PID reuse."
                )
                LOG.error(error)
                raise ValueError(error)

            # Check 2: Command line contains expected tokens
            cmdline = process.cmdline()
            cmdline_str = " ".join(cmdline)

            tokens_found = [
                token for token in expected_cmdline_tokens
                if token.lower() in cmdline_str.lower()
            ]

            if not tokens_found:
                error = (
                    f"PID {pid} cmdline does not contain expected tokens {expected_cmdline_tokens}. "
                    f"Command: {' '.join(cmdline[:3])}... Possible PID reuse."
                )
                LOG.error(error)
                raise ValueError(error)

            # Check 3: Recent creation warning
            if warn_recent > 0:
                import time
                create_time = process.create_time()
                if time.time() - create_time < warn_recent:
                    LOG.warning(
                        f"PID {pid} created very recently ({time.time() - create_time:.2f}s ago). "
                        f"Possible PID reuse - verify this is correct process."
                    )

            LOG.debug(f"Verified PID {pid} matches expected process")

            return ProcessVerification(
                pid=pid,
                is_verified=True,
                process_name=process.name(),
                cmdline=cmdline_str
            )

        except psutil.NoSuchProcess:
            raise ValueError(f"PID {pid} does not exist")

        except psutil.AccessDenied:
            raise ValueError(f"Access denied when checking PID {pid}")

    else:
        # Fallback: Check /proc/{pid}/cmdline
        try:
            with open(f"/proc/{pid}/cmdline", "rb") as f:
                cmdline_bytes = f.read()
                cmdline = cmdline_bytes.decode("utf-8", errors="ignore")

            # Check name (approximate - cmdline has first arg)
            name_match = any(name.lower() in cmdline.lower() for name in expected_names)
            if not name_match:
                error = f"PID {pid} cmdline does not match expected names {expected_names}"
                LOG.error(error)
                raise ValueError(error)

            # Check tokens
            tokens_found = [
                token for token in expected_cmdline_tokens
                if token.lower() in cmdline.lower()
            ]

            if not tokens_found:
                error = f"PID {pid} cmdline does not contain expected tokens {expected_cmdline_tokens}"
                LOG.error(error)
                raise ValueError(error)

            LOG.debug(f"Verified PID {pid} matches expected process (via /proc)")

            return ProcessVerification(
                pid=pid,
                is_verified=True,
                process_name=cmdline.split('\0')[0] if '\0' in cmdline else cmdline[:50],
                cmdline=cmdline
            )

        except FileNotFoundError:
            raise ValueError(f"PID {pid} does not exist")



"""
QMP Module - Future Enhancements Placeholder

This module is reserved for future QMP abstractions and utilities.
Currently, QMP functionality is implemented in:
  - maqet/machine.py: Core QMP command execution
  - maqet/qmp/keyboard.py: Keyboard emulation utilities

Future enhancements to consider:
  - QMP command abstractions and builder pattern
  - QMP event handling and monitoring
  - Async QMP support for non-blocking operations
  - QMP capabilities negotiation
  - QMP command queueing and rate limiting
  - QMP response validation and error handling
  - High-level QMP abstractions (device management, migration, etc.)
"""



"""
API Registry provider abstraction for testability.

Separates "what" (registry) from "how" (initialization).
This enables different initialization strategies for production vs testing:
- Production: singleton global registry
- Testing: worker-scoped registry for parallel test isolation
"""

from typing import Protocol, Optional
from maqet.api.registry import APIRegistry


class APIRegistryProvider(Protocol):
    """Protocol for providing API registry instances."""

    def get_registry(self) -> APIRegistry:
        """Get or create registry for current context."""
        ...


class ProductionRegistryProvider:
    """Production provider: singleton global registry."""

    def get_registry(self) -> APIRegistry:
        """Return global singleton registry."""
        from maqet.api import API_REGISTRY
        return API_REGISTRY


class WorkerScopedRegistryProvider:
    """Test provider: one registry per worker process."""

    def __init__(self):
        """Initialize provider with no registry."""
        self._registry: Optional[APIRegistry] = None

    def get_registry(self) -> APIRegistry:
        """Get or create worker-scoped registry."""
        if self._registry is None:
            import inspect
            from maqet.maqet import Maqet
            from maqet.api.metadata import APIMethodMetadata

            self._registry = APIRegistry()

            # Register Maqet methods to this specific registry instance
            for name, method in inspect.getmembers(Maqet, predicate=inspect.isfunction):
                if hasattr(method, "_api_metadata"):
                    metadata: APIMethodMetadata = method._api_metadata
                    # Set the owner class now that we know it
                    metadata.owner_class = Maqet.__name__
                    # Register with our worker-scoped registry
                    self._registry.register(metadata)

        return self._registry



"""
Configuration Merger

Handles deep-merging of multiple configuration files.
Provides utilities for loading and merging YAML configuration files
with deep merge support for complex nested structures.
"""

from pathlib import Path
from typing import Any, Dict, List, Union

import yaml

from ..exceptions import ConfigurationError


class ConfigLimits:
    """Configuration protection limits to prevent DoS attacks."""
    MAX_DEPTH = 10
    MAX_SIZE_BYTES = 1 * 1024 * 1024  # 1MB
    MAX_KEYS_PER_LEVEL = 1000


class ConfigError(ConfigurationError):
    """Configuration parsing errors (alias for ConfigurationError)."""


class ConfigMergeError(ConfigError):
    """Error during configuration merge (depth/size limits exceeded)."""


class ConfigMerger:
    """
    Handles deep-merging of multiple configuration files.

    Provides utilities for loading and merging YAML configuration files
    with deep merge support for complex nested structures.
    """

    @staticmethod
    def _merge_arguments_list(
        base_args: List[Any], override_args: List[Any]
    ) -> List[Any]:
        """
        Merge two arguments lists with override behavior.

        For arguments like [{foo: 0}, {bar: 10}] and [{bar: 20}, {baz: 30}],
        later configs override earlier ones by key, producing:
        [{foo: 0}, {bar: 20}, {baz: 30}]

        Args:
            base_args: Base arguments list
            override_args: Override arguments list

        Returns:
            Merged arguments list with overrides applied
        """
        # Track arguments by their keys (for dict items)
        # Use dict to maintain insertion order (Python 3.7+)
        merged = {}

        # Process base arguments first
        for arg in base_args:
            if isinstance(arg, dict):
                # For dict items, use the key as identifier
                for key in arg.keys():
                    merged[key] = arg
            else:
                # For non-dict items (strings), use the item itself as key
                merged[str(arg)] = arg

        # Process override arguments (later configs win)
        for arg in override_args:
            if isinstance(arg, dict):
                # Override or add dict items by key
                for key in arg.keys():
                    merged[key] = arg
            else:
                # Override or add non-dict items
                merged[str(arg)] = arg

        # Convert back to list, preserving order
        return list(merged.values())

    @staticmethod
    def deep_merge(
        base: Dict[str, Any],
        override: Dict[str, Any],
        *,
        _depth: int = 0,
        max_depth: int = ConfigLimits.MAX_DEPTH,
    ) -> Dict[str, Any]:
        """
        Deep merge two dictionaries with protection limits.

        Args:
            base: Base configuration dictionary
            override: Override configuration dictionary
            _depth: Current recursion depth (internal use)
            max_depth: Maximum allowed recursion depth

        Returns:
            Deep-merged configuration dictionary

        Raises:
            ConfigMergeError: If depth or size limits exceeded
        """
        # Check depth limit
        if _depth > max_depth:
            raise ConfigMergeError(
                f"Configuration nesting exceeds maximum depth ({max_depth}). "
                "This may indicate a circular reference or malformed config."
            )

        # Check key count at this level
        if len(base) > ConfigLimits.MAX_KEYS_PER_LEVEL:
            raise ConfigMergeError(
                f"Configuration has too many keys at depth {_depth} "
                f"({len(base)} > {ConfigLimits.MAX_KEYS_PER_LEVEL})"
            )

        result = base.copy()

        for key, value in override.items():
            if (
                key in result
                and isinstance(result[key], dict)
                and isinstance(value, dict)
            ):
                # Recursively merge nested dictionaries with depth tracking
                result[key] = ConfigMerger.deep_merge(
                    result[key], value, _depth=_depth + 1, max_depth=max_depth
                )
            elif (
                key in result
                and isinstance(result[key], list)
                and isinstance(value, list)
            ):
                # Special handling for 'arguments' list: merge dict items by key
                if key == "arguments":
                    result[key] = ConfigMerger._merge_arguments_list(
                        result[key], value
                    )
                else:
                    # For other lists (storage, etc.), concatenate
                    # This allows adding storage devices, network interfaces, etc.
                    result[key] = result[key] + value
            else:
                # Override scalar values and non-dict types
                result[key] = value

        return result

    # QEMU arguments that accept file paths as direct values
    # These will have their relative paths resolved to absolute
    PATH_BEARING_ARGS = frozenset({
        'kernel',     # Kernel image path
        'initrd',     # Initial ramdisk path
        'bios',       # BIOS/UEFI firmware path
        'pflash',     # Flash ROM images (can have multiple)
        'dtb',        # Device Tree Blob path
        'rom',        # ROM file path (option-rom)
        'loadvm',     # VM snapshot file
        'incoming',   # Migration file (can be file:path)
    })

    # Pattern prefixes in argument values that indicate file paths
    # Format: "prefix=path" or "prefix:path"
    PATH_PATTERNS = frozenset({
        'file=',      # drive file=./disk.qcow2
        'path=',      # chardev path=/tmp/serial
        'unix:',      # vnc unix:/tmp/vnc.sock
        'socket=',    # spice socket=/tmp/spice.sock
    })

    @staticmethod
    def _resolve_path_value(value: str, config_dir: Path) -> str:
        """
        Resolve a single path value if it's relative.

        Args:
            value: Path string (may be relative or absolute)
            config_dir: Directory to resolve relative paths against

        Returns:
            Resolved absolute path string
        """
        if not value or not isinstance(value, str):
            return value

        # Check if value is a path and relative
        value_path = Path(value)
        if not value_path.is_absolute():
            return str((config_dir / value).resolve())

        return value

    @staticmethod
    def _resolve_pattern_in_value(
        value: str, config_dir: Path
    ) -> str:
        """
        Resolve file paths in complex argument values with patterns.

        Handles patterns like:
        - file=/path/to/disk.qcow2,media=cdrom
        - path=/tmp/serial,mode=0600
        - unix:/tmp/vnc.sock

        Args:
            value: Argument value string
            config_dir: Directory to resolve relative paths against

        Returns:
            Value with resolved paths
        """
        if not isinstance(value, str):
            return value

        # Check for pattern-based paths
        for pattern in ConfigMerger.PATH_PATTERNS:
            if pattern in value:
                if ',' in value:
                    # Complex pattern: file=./disk.qcow2,media=cdrom
                    parts = value.split(",")
                    for i, part in enumerate(parts):
                        if part.startswith(pattern):
                            # Extract path after pattern prefix
                            path_value = part[len(pattern):]
                            if path_value and not Path(path_value).is_absolute():
                                abs_path = str(
                                    (config_dir / path_value).resolve()
                                )
                                parts[i] = f"{pattern}{abs_path}"
                    return ",".join(parts)
                elif value.startswith(pattern):
                    # Simple pattern: file=./disk.qcow2 or unix:/tmp/vnc.sock
                    path_value = value[len(pattern):]
                    if path_value and not Path(path_value).is_absolute():
                        abs_path = str(
                            (config_dir / path_value).resolve()
                        )
                        return f"{pattern}{abs_path}"

        return value

    @staticmethod
    def _resolve_relative_paths(
        config_data: Dict[str, Any], config_dir: Path
    ) -> Dict[str, Any]:
        """
        Resolve relative file paths in config to absolute paths.

        Layer 3 of automatic path resolution: resolve paths relative to config file location.
        This makes configs portable and allows relative paths like ./live.iso to work.

        Resolves paths in:
        1. storage[].file - Main storage file path (QCOW2, raw, etc.)
        2. storage[].path - VirtFS shared directory path
        3. Path-bearing arguments (kernel, initrd, bios, pflash, dtb, rom, etc.)
        4. Pattern-based paths in arguments (file=, path=, unix:, socket=)

        Note: storage[].backing_file is NOT resolved here. It's resolved relative
        to the delta image directory by storage.py (see commit a6cf3e6).

        Args:
            config_data: Configuration dictionary
            config_dir: Directory containing the config file

        Returns:
            Configuration with resolved absolute paths
        """
        # Resolve storage device file paths using helper method
        if "storage" in config_data and isinstance(
            config_data["storage"], list
        ):
            for storage_item in config_data["storage"]:
                if isinstance(storage_item, dict):
                    # Resolve main storage file path
                    if "file" in storage_item:
                        storage_item["file"] = ConfigMerger._resolve_path_value(
                            storage_item["file"], config_dir
                        )

                    # DON'T resolve backing_file here!
                    # backing_file paths are resolved relative to delta image directory
                    # by storage.py, not config directory. See commit a6cf3e6.
                    # This enables portable delta+base image pairs that can be moved together.

                    # Resolve VirtFS path entries
                    if "path" in storage_item:
                        storage_item["path"] = ConfigMerger._resolve_path_value(
                            storage_item["path"], config_dir
                        )

        # Resolve file paths in arguments
        if "arguments" in config_data and isinstance(
            config_data["arguments"], list
        ):
            for arg_item in config_data["arguments"]:
                if isinstance(arg_item, dict):
                    for key, value in list(arg_item.items()):
                        # Skip non-string values
                        if not isinstance(value, str):
                            continue

                        # Pattern 1: Direct path-bearing arguments
                        # Example: {kernel: "./vmlinuz"} -> {kernel: "/abs/path/vmlinuz"}
                        if key in ConfigMerger.PATH_BEARING_ARGS:
                            arg_item[key] = ConfigMerger._resolve_path_value(
                                value, config_dir
                            )
                            continue

                        # Pattern 2: Complex values with path patterns
                        # Example: {drive: "file=./disk.qcow2,media=cdrom"}
                        # Only process if value contains a pattern to avoid unnecessary overhead
                        if any(pattern in value for pattern in ConfigMerger.PATH_PATTERNS):
                            arg_item[key] = ConfigMerger._resolve_pattern_in_value(
                                value, config_dir
                            )

        return config_data

    @staticmethod
    def validate_config_size(config_path: Path) -> None:
        """Validate config file size before loading.

        Args:
            config_path: Path to configuration file

        Raises:
            ConfigMergeError: If file exceeds size limit
        """
        size = config_path.stat().st_size
        if size > ConfigLimits.MAX_SIZE_BYTES:
            raise ConfigMergeError(
                f"Configuration file too large: {size} bytes "
                f"(max: {ConfigLimits.MAX_SIZE_BYTES})"
            )

    @staticmethod
    def load_and_merge_files(
        config_files: Union[str, List[str]],
    ) -> Dict[str, Any]:
        """
        Load and merge multiple configuration files.

        Args:
            config_files: Single config file path or list of config file paths

        Returns:
            Merged configuration data

        Raises:
            ConfigError: If any config file cannot be loaded or parsed
        """
        if isinstance(config_files, str):
            config_files = [config_files]

        if not config_files:
            return {}

        merged_config = {}

        for config_file in config_files:
            config_path = Path(config_file)
            if not config_path.exists():
                raise ConfigError(
                    f"Configuration file not found: {config_file}"
                )

            # Validate file size before loading
            ConfigMerger.validate_config_size(config_path)

            try:
                with open(config_path) as f:
                    config_data = yaml.safe_load(f) or {}

                if not isinstance(config_data, dict):
                    raise ConfigError(
                        f"Configuration in {config_file} must be a "
                        f"YAML dictionary"
                    )

                # Layer 3: Resolve relative paths in config relative to config file location
                config_dir = config_path.parent.resolve()
                config_data = ConfigMerger._resolve_relative_paths(
                    config_data, config_dir
                )

                # Deep merge with previous configs
                merged_config = ConfigMerger.deep_merge(
                    merged_config, config_data
                )

            except yaml.YAMLError as e:
                raise ConfigError(f"Invalid YAML in {config_file}: {e}")
            except Exception as e:
                raise ConfigError(
                    f"Error loading configuration from {config_file}: {e}"
                )

        return merged_config



"""
Configuration Manager

Centralized configuration management for maqet.
Consolidates configuration from CLI args, environment variables,
config files, and defaults with clear precedence rules.
"""

from pathlib import Path
from threading import RLock
from typing import Any, Dict, Optional

from ..logger import LOG
from .merger import ConfigError, ConfigMerger
from .parser import ConfigParser
from .runtime_config import RuntimeConfig


class ConfigManager:
    """
    Centralized configuration management for maqet.

    Consolidates configuration from CLI args, environment variables,
    config files, and defaults with clear precedence rules.

    Precedence (highest to lowest):
    1. CLI arguments (cli_overrides)
    2. Environment variables (via RuntimeConfig)
    3. Config files (maqet.conf, VM YAML configs)
    4. Defaults (RuntimeConfig.DEFAULTS)

    Example:
        # Create with CLI overrides
        config = ConfigManager(cli_overrides={
            "directories": {"data_dir": "/custom/data"}
        })

        # Access configuration
        data_dir = config.get_data_dir()
        memory = config.get_vm_default("memory", "2G")
        verbosity = config.get_verbosity()
    """

    def __init__(
        self,
        cli_overrides: Optional[Dict[str, Any]] = None,
        runtime_config: Optional[RuntimeConfig] = None,
        config_parser: Optional[ConfigParser] = None,
    ):
        """
        Initialize configuration manager.

        Args:
            cli_overrides: CLI arguments that override all other sources
            runtime_config: Pre-loaded RuntimeConfig instance (optional)
            config_parser: ConfigParser instance for VM configs (optional)

        Example:
            config = ConfigManager(cli_overrides={
                "directories": {"data_dir": "/custom/data"},
                "logging": {"verbosity": 2}
            })
        """
        self.cli_overrides = cli_overrides or {}
        self._runtime_config = runtime_config
        self._config_parser = config_parser
        self._config: Dict[str, Any] = {}
        self._lock = RLock()

        # Load and merge all configurations
        self.load()

    def load(self) -> Dict[str, Any]:
        """
        Load and merge all configurations.

        Precedence: CLI > Environment > Config File > Defaults

        Returns:
            Merged configuration dictionary
        """
        with self._lock:
            # Start with defaults from RuntimeConfig
            config = RuntimeConfig.DEFAULTS.copy()

            # Load runtime config (includes env vars and config files)
            if self._runtime_config is None:
                self._runtime_config = RuntimeConfig()

            # Merge runtime config (access internal _config)
            runtime_data = self._runtime_config._config
            config = ConfigMerger.deep_merge(config, runtime_data)

            # Merge CLI overrides (highest priority)
            config = ConfigMerger.deep_merge(config, self.cli_overrides)

            # Validate configuration
            self._validate(config)

            self._config = config
            return config

    def reload(self) -> None:
        """
        Reload configuration from disk.

        Useful for long-running processes that need to pick up config changes.
        Resets cache and re-loads all configuration sources.
        """
        with self._lock:
            # Reset runtime config to force reload
            self._runtime_config = None
            self.load()
            LOG.info("Configuration reloaded from disk")

    def get(self, key: str, default: Any = None) -> Any:
        """
        Get configuration value using dot notation.

        Args:
            key: Dot-separated key path (e.g., "directories.data_dir")
            default: Default value if key not found

        Returns:
            Configuration value or default

        Example:
            data_dir = config.get("directories.data_dir")
            memory = config.get("vm_defaults.memory", "2G")
        """
        with self._lock:
            keys = key.split(".")
            value = self._config

            for k in keys:
                if isinstance(value, dict) and k in value:
                    value = value[k]
                else:
                    return default

            return value

    def get_data_dir(self) -> Path:
        """
        Get data directory path.

        Returns:
            Data directory path (guaranteed to return valid Path)
        """
        data_dir = self.get("directories.data_dir")
        if data_dir:
            return Path(data_dir)

        # Fallback to RuntimeConfig, then XDG default
        runtime_dir = self._runtime_config.get_data_dir()
        if runtime_dir:
            return Path(runtime_dir)

        # Ultimate fallback: XDG standard
        import os

        xdg_data_home = os.environ.get(
            "XDG_DATA_HOME", str(Path.home() / ".local" / "share")
        )
        return Path(xdg_data_home) / "maqet"

    def get_config_dir(self) -> Path:
        """
        Get config directory path.

        Returns:
            Config directory path (guaranteed to return valid Path)
        """
        config_dir = self.get("directories.config_dir")
        if config_dir:
            return Path(config_dir)

        # Fallback to RuntimeConfig, then XDG default
        runtime_dir = self._runtime_config.get_config_dir()
        if runtime_dir:
            return Path(runtime_dir)

        # Ultimate fallback: XDG standard
        import os

        xdg_config_home = os.environ.get(
            "XDG_CONFIG_HOME", str(Path.home() / ".config")
        )
        return Path(xdg_config_home) / "maqet"

    def get_runtime_dir(self) -> Path:
        """
        Get runtime directory path.

        Returns:
            Runtime directory path (guaranteed to return valid Path)
        """
        runtime_dir = self.get("directories.runtime_dir")
        if runtime_dir:
            return Path(runtime_dir)

        # Fallback to RuntimeConfig, then XDG default
        rt_dir = self._runtime_config.get_runtime_dir()
        if rt_dir:
            return Path(rt_dir)

        # Ultimate fallback: XDG standard
        import os

        xdg_runtime_dir = os.environ.get("XDG_RUNTIME_DIR")
        if xdg_runtime_dir:
            return Path(xdg_runtime_dir) / "maqet"

        # Fallback if XDG_RUNTIME_DIR not set
        return Path("/tmp") / f"maqet-{os.getuid()}"

    def get_vm_default(self, key: str, default: Any = None) -> Any:
        """
        Get VM default configuration value.

        Args:
            key: VM config key (e.g., "memory", "cpu")
            default: Default value if not found

        Returns:
            VM default value or default

        Example:
            memory = config.get_vm_default("memory", "2G")
            cpu = config.get_vm_default("cpu", 2)
        """
        return self.get(f"vm_defaults.{key}", default)

    def get_verbosity(self) -> int:
        """
        Get logging verbosity level.

        Returns:
            Verbosity level (0-3):
            - 0: Errors only
            - 1: Warnings
            - 2: Info
            - 3: Debug
        """
        return self.get("logging.verbosity", 0)

    def get_log_file(self) -> Optional[Path]:
        """
        Get log file path.

        Returns:
            Log file path or None if not configured
        """
        log_file = self.get("logging.log_file")
        return Path(log_file) if log_file else None

    def validate(self) -> None:
        """
        Validate all configuration values.

        Raises:
            ConfigError: If any configuration value is invalid
        """
        with self._lock:
            self._validate(self._config)

    def validate_vm_config(self, config_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Validate VM-specific configuration.

        Args:
            config_data: VM configuration dictionary

        Returns:
            Validated configuration data

        Raises:
            ConfigValidationError: If validation fails

        Example:
            config = {
                'binary': '/usr/bin/qemu-system-x86_64',
                'memory': '4G',
                'cpu': 2,
                'storage': [
                    {'name': 'hdd', 'size': '20G', 'type': 'qcow2'}
                ]
            }
            validated = config_manager.validate_vm_config(config)
        """
        # Use ConfigParser for VM validation if available
        if self._config_parser:
            return self._config_parser.validate_config(config_data)

        # Otherwise use validators module directly
        from .schema_validators import validate_config_data

        return validate_config_data(config_data)

    def to_dict(self) -> Dict[str, Any]:
        """
        Export configuration as dictionary.

        Returns:
            Complete configuration dictionary (deep copy)
        """
        with self._lock:
            import copy

            return copy.deepcopy(self._config)

    def get_config_sources(self) -> Dict[str, Any]:
        """
        Get information about configuration sources.

        Returns:
            Dictionary with:
            - config_file_path: Path to loaded config file
            - cli_overrides: CLI arguments applied
            - defaults_used: List of keys using default values

        Example:
            sources = config.get_config_sources()
            print(f"Config file: {sources['config_file_path']}")
            print(f"CLI overrides: {sources['cli_overrides']}")
        """
        with self._lock:
            config_file_path = None
            if self._runtime_config:
                config_file_path = self._runtime_config._config_file_path

            # Find keys using default values
            defaults_used = []
            for key in self._flatten_keys(RuntimeConfig.DEFAULTS):
                if self.get(key) == self._get_from_dict(
                    RuntimeConfig.DEFAULTS, key
                ):
                    defaults_used.append(key)

            return {
                "config_file_path": str(config_file_path)
                if config_file_path
                else None,
                "cli_overrides": self.cli_overrides.copy(),
                "defaults_used": defaults_used,
            }

    def _validate(self, config: Dict[str, Any]) -> None:
        """
        Validate configuration values.

        Args:
            config: Configuration dictionary to validate

        Raises:
            ConfigError: If validation fails
        """
        # Validate verbosity
        verbosity = config.get("logging", {}).get("verbosity")
        if verbosity is not None:
            if not isinstance(verbosity, int):
                raise ConfigError(
                    f"Invalid verbosity type: {type(verbosity).__name__}. "
                    f"Must be int (0-3)"
                )
            if verbosity < 0 or verbosity > 3:
                raise ConfigError(
                    f"Invalid verbosity: {verbosity}. Must be 0-3 "
                    f"(0=errors, 1=warnings, 2=info, 3=debug)"
                )

        # Validate log file path
        log_file = config.get("logging", {}).get("log_file")
        if log_file:
            log_path = Path(log_file)
            log_dir = log_path.parent
            if not log_dir.exists():
                raise ConfigError(
                    f"Log file directory does not exist: {log_dir}"
                )

        # Validate VM defaults if present
        vm_defaults = config.get("vm_defaults", {})
        if vm_defaults:
            # Validate memory format
            memory = vm_defaults.get("memory")
            if memory and not self._is_valid_memory_format(memory):
                raise ConfigError(
                    f"Invalid memory format: {memory}. "
                    f"Use format like '2G', '512M', '4096K'"
                )

            # Validate CPU count
            cpu = vm_defaults.get("cpu")
            if cpu is not None:
                if not isinstance(cpu, int):
                    raise ConfigError(
                        f"Invalid CPU type: {type(cpu).__name__}. Must be int"
                    )
                if cpu < 1:
                    raise ConfigError(
                        f"Invalid CPU count: {cpu}. Must be >= 1"
                    )

    def _is_valid_memory_format(self, memory: str) -> bool:
        """
        Validate memory format.

        Args:
            memory: Memory string (e.g., '2G', '512M')

        Returns:
            True if valid format, False otherwise
        """
        import re

        return bool(re.match(r"^\d+[KMGT]$", str(memory)))

    def _flatten_keys(self, d: Dict[str, Any], prefix: str = "") -> list:
        """
        Flatten nested dictionary keys into dot notation.

        Args:
            d: Dictionary to flatten
            prefix: Key prefix for recursion

        Returns:
            List of flattened keys
        """
        keys = []
        for k, v in d.items():
            new_key = f"{prefix}.{k}" if prefix else k
            if isinstance(v, dict):
                keys.extend(self._flatten_keys(v, new_key))
            else:
                keys.append(new_key)
        return keys

    def _get_from_dict(self, d: Dict[str, Any], key: str) -> Any:
        """
        Get value from nested dictionary using dot notation.

        Args:
            d: Dictionary to search
            key: Dot-separated key path

        Returns:
            Value or None if not found
        """
        keys = key.split(".")
        value = d
        for k in keys:
            if isinstance(value, dict) and k in value:
                value = value[k]
            else:
                return None
        return value



"""
VM Repository

Handles CRUD operations for VM instances in the SQLite database.
Extracted from StateManager to follow Single Responsibility Principle.
"""

import json
import re
import sqlite3
import time
import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

if TYPE_CHECKING:
    from ..state import VMInstance

from ..constants import Intervals, Retries
from ..exceptions import StateError
from ..logger import LOG

# Optional dependency - check availability without importing
try:
    from importlib.util import find_spec
    PSUTIL_AVAILABLE = find_spec("psutil") is not None
except ImportError:
    PSUTIL_AVAILABLE = False


def _calculate_retry_delay(attempt: int, base_delay: float = 0.1, max_delay: float = 2.0) -> float:
    """Calculate retry delay with exponential backoff and decorrelated jitter.

    Uses decorrelated jitter (AWS recommended) to prevent thundering herd
    while strictly respecting max_delay. The delay is randomized between
    base_delay and the exponential backoff value, ensuring it never exceeds
    max_delay.

    Args:
        attempt: Current attempt number (0-indexed)
        base_delay: Base delay in seconds
        max_delay: Maximum delay cap in seconds

    Returns:
        Delay in seconds with jitter applied (always <= max_delay)
    """
    import random
    # Exponential backoff: base * 2^attempt, capped at max_delay
    exponential = min(base_delay * (2 ** attempt), max_delay)
    # Decorrelated jitter: random between base_delay and exponential
    # This prevents thundering herd while respecting max_delay
    return random.uniform(base_delay, exponential)


class VMRepository:
    """Repository for VM instance CRUD operations."""

    def __init__(self, get_connection: callable, get_pooled_connection: callable) -> None:
        """
        Initialize VM Repository.

        Args:
            get_connection: Callable that returns a context manager yielding sqlite3.Connection
            get_pooled_connection: Callable that returns a context manager yielding pooled connection
        """
        self._get_connection = get_connection
        self._get_pooled_connection = get_pooled_connection

    def create_vm(
        self,
        name: str,
        config_data: Dict[str, Any],
        config_path: Optional[str] = None,
        get_vm_callback: callable = None,
    ) -> str:
        """
        Create a new VM instance.

        Args:
            name: VM name (must be unique)
            config_data: VM configuration dictionary
            config_path: Optional path to config file
            get_vm_callback: Callback to check if VM exists (to avoid circular dependency)

        Returns:
            VM instance ID

        Raises:
            StateError: If validation fails or DB operation fails
        """
        # Validation: Check VM name
        if not name or not name.strip():
            raise StateError("VM name cannot be empty")

        # Check name length (prevent filesystem issues with very long names)
        if len(name) > 255:
            raise StateError(
                f"VM name too long ({len(name)} chars). Maximum is 255 characters."
            )

        # Validate name contains only safe characters (alphanumeric, dash,
        # underscore, dot)
        if not re.match(r"^[a-zA-Z0-9._-]+$", name):
            raise StateError(
                f"VM name '{name}' contains invalid characters. "
                f"Only alphanumeric, dash (-), underscore (_), and dot (.) are allowed."
            )

        # Check for name conflicts BEFORE attempting insert
        # Provides clearer error message than SQLite IntegrityError
        if get_vm_callback:
            existing_vm = get_vm_callback(name)
            if existing_vm:
                raise StateError(
                    f"VM with name '{name}' already exists (ID: {existing_vm.id}). "
                    f"Use 'maqet rm {name}' to remove it first, "
                    f"or choose a different name."
                )

        # Validation: Check config_data size to prevent DB bloat
        config_json = json.dumps(config_data)
        config_size = len(config_json.encode("utf-8"))
        max_config_size = 10 * 1024 * 1024  # 10MB limit

        if config_size > max_config_size:
            raise StateError(
                f"Configuration data too large ({config_size} bytes, max {max_config_size}). "
                f"Consider reducing storage device configurations or using external config files."
            )

        vm_id = str(uuid.uuid4())

        # NOTE: auth_secret no longer stored in database (ephemeral file-based secrets)
        # Secret is generated by VMRunner at runtime and stored in {socket_path}.auth

        # Retry logic for transient lock errors
        max_retries = Retries.DB_OPERATION
        for attempt in range(max_retries):
            try:
                with self._get_pooled_connection(readonly=False) as conn:
                    conn.execute(
                        """
                        INSERT INTO vm_instances (id, name, config_path, config_data, status)
                        VALUES (?, ?, ?, ?, 'created')
                    """,
                        (vm_id, name, config_path, config_json),
                    )
                return vm_id  # Success
            except sqlite3.IntegrityError:
                raise StateError(f"VM with name '{name}' already exists")
            except sqlite3.OperationalError as e:
                error_msg = str(e).lower()

                # Handle transient lock errors with retry
                if "locked" in error_msg and attempt < max_retries - 1:
                    # Exponential backoff with jitter
                    wait_time = _calculate_retry_delay(attempt, Intervals.DB_RETRY_BASE)
                    LOG.debug(
                        f"Database locked while creating VM (attempt {attempt + 1}/{max_retries}), "
                        f"retrying in {wait_time:.2f}s"
                    )
                    time.sleep(wait_time)
                    continue

                # Non-retryable or final attempt - raise appropriate error
                if "disk" in error_msg or "space" in error_msg:
                    raise StateError(
                        f"Cannot create VM: Disk full or insufficient space. "
                        f"Free up disk space and try again. Error: {e}"
                    )
                elif "locked" in error_msg:
                    raise StateError(
                        f"Failed to create VM after {max_retries} retries: Database is locked. "
                        f"Another process may be accessing it. Error: {e}"
                    )
                else:
                    raise StateError(
                        f"Database error while creating VM: {e}"
                    )
            except Exception as e:
                # Log unexpected errors with context
                LOG.error(
                    f"Unexpected error creating VM '{name}' (ID: {vm_id}): {type(e).__name__}: {e}"
                )
                raise StateError(f"Failed to create VM: {e}")

        # Should not reach here due to return in success case or exception
        raise StateError(f"Failed to create VM after {max_retries} retries")

    def get_vm(self, identifier: str) -> Optional["VMInstance"]:
        """
        Get VM instance by ID or name.

        Optimized to use indexes by trying ID lookup first (PRIMARY KEY),
        then name lookup (idx_vm_name index) if not found.

        Args:
            identifier: VM ID or name

        Returns:
            VM instance or None if not found
        """
        # NOTE: SECURITY - Uses parameterized queries, safe from SQL injection
        with self._get_pooled_connection(readonly=True) as conn:
            # Try ID first (PRIMARY KEY index - O(log n))
            row = conn.execute(
                "SELECT * FROM vm_instances WHERE id = ?",
                (identifier,),
            ).fetchone()

            if row:
                return self._row_to_vm_instance(row)

            # Try name (idx_vm_name index - O(log n))
            row = conn.execute(
                "SELECT * FROM vm_instances WHERE name = ?",
                (identifier,),
            ).fetchone()

            if row:
                return self._row_to_vm_instance(row)

        return None

    def list_vms(
        self, status_filter: Optional[str] = None
    ) -> List["VMInstance"]:
        """
        List all VM instances.

        Args:
            status_filter: Optional status to filter by

        Returns:
            List of VM instances
        """
        with self._get_pooled_connection(readonly=True) as conn:
            if status_filter:
                rows = conn.execute(
                    "SELECT * FROM vm_instances WHERE status = ? ORDER BY created_at",
                    (status_filter,),
                ).fetchall()
            else:
                rows = conn.execute(
                    "SELECT * FROM vm_instances ORDER BY created_at"
                ).fetchall()

            return [self._row_to_vm_instance(row) for row in rows]

    def update_vm_status(
        self,
        identifier: str,
        status: str,
        pid: Optional[int] = None,
        runner_pid: Optional[int] = None,
        socket_path: Optional[str] = None,
        qmp_socket_path: Optional[str] = None,
        validate_pid_callback: Optional[callable] = None,
    ) -> bool:
        """
        Update VM status and process information.

        Optimized to use indexes by trying ID lookup first (PRIMARY KEY),
        then name lookup (idx_vm_name index) if not found.

        Args:
            identifier: VM ID or name
            status: New status
            pid: QEMU process ID (if running)
            runner_pid: VM runner process ID (if running)
            socket_path: QMP socket path (if running)
            qmp_socket_path: QMP socket path for cross-process communication
            validate_pid_callback: Callback to validate PID ownership (security concern)

        Returns:
            True if updated, False if VM not found

        Raises:
            ValueError: If PID ownership validation fails
        """
        # Security: Validate PID ownership if provided and callback available
        if validate_pid_callback:
            if pid is not None:
                try:
                    validate_pid_callback(pid)
                except ValueError as e:
                    LOG.error(f"PID validation failed: {e}")
                    raise

            if runner_pid is not None:
                try:
                    validate_pid_callback(runner_pid)
                except ValueError as e:
                    LOG.error(f"Runner PID validation failed: {e}")
                    raise

        # Retry logic for transient lock errors
        max_retries = Retries.DB_OPERATION
        for attempt in range(max_retries):
            try:
                with self._get_pooled_connection(readonly=False) as conn:
                    # Try ID first (PRIMARY KEY index - O(log n))
                    cursor = conn.execute(
                        """
                        UPDATE vm_instances
                        SET status = ?, pid = ?, runner_pid = ?, socket_path = ?, qmp_socket_path = ?
                        WHERE id = ?
                    """,
                        (status, pid, runner_pid, socket_path, qmp_socket_path, identifier),
                    )

                    if cursor.rowcount > 0:
                        return True

                    # Try name (idx_vm_name index - O(log n))
                    cursor = conn.execute(
                        """
                        UPDATE vm_instances
                        SET status = ?, pid = ?, runner_pid = ?, socket_path = ?, qmp_socket_path = ?
                        WHERE name = ?
                    """,
                        (status, pid, runner_pid, socket_path, qmp_socket_path, identifier),
                    )

                    return cursor.rowcount > 0

            except sqlite3.OperationalError as e:
                if "locked" in str(e).lower() and attempt < max_retries - 1:
                    # Exponential backoff with jitter
                    wait_time = _calculate_retry_delay(attempt, Intervals.DB_RETRY_BASE)
                    LOG.debug(
                        f"Database locked while updating VM status (attempt {attempt + 1}/{max_retries}), "
                        f"retrying in {wait_time:.2f}s"
                    )
                    time.sleep(wait_time)
                    continue
                raise StateError(f"Failed to update VM status after {max_retries} retries: {e}")

        # Should not reach here
        return False

    def batch_update_vm_statuses(
        self,
        updates: Dict[str, Tuple[str, Optional[int], Optional[int], Optional[str]]]
    ) -> int:
        """Update multiple VM statuses in a single transaction.

        Efficiently updates status, pid, runner_pid, and socket_path for multiple VMs
        in one database transaction, avoiding N+1 query pattern.

        Args:
            updates: Dict mapping VM name to tuple of (status, pid, runner_pid, socket_path)

        Returns:
            Number of rows updated

        Raises:
            StateError: If database operation fails
        """
        if not updates:
            return 0

        max_retries = Retries.DB_OPERATION
        for attempt in range(max_retries):
            try:
                with self._get_pooled_connection(readonly=False) as conn:
                    # Use executemany for batch update (single transaction)
                    conn.executemany(
                        """
                        UPDATE vm_instances
                        SET status = ?, pid = ?, runner_pid = ?, socket_path = ?
                        WHERE name = ?
                        """,
                        [
                            (status, pid, runner_pid, socket_path, name)
                            for name, (status, pid, runner_pid, socket_path) in updates.items()
                        ]
                    )
                    LOG.debug(f"Batch updated {len(updates)} VM status(es)")
                    return len(updates)

            except sqlite3.OperationalError as e:
                if "locked" in str(e).lower() and attempt < max_retries - 1:
                    # Exponential backoff with jitter
                    wait_time = _calculate_retry_delay(attempt, Intervals.DB_RETRY_BASE)
                    LOG.debug(
                        f"Database locked during batch update (attempt {attempt + 1}/{max_retries}), "
                        f"retrying in {wait_time:.2f}s"
                    )
                    time.sleep(wait_time)
                    continue
                raise StateError(
                    f"Failed to batch update VM statuses after {max_retries} retries: {e}"
                )

        return 0

    def remove_vm(self, identifier: str) -> bool:
        """
        Remove VM instance from database.

        Optimized to use indexes by trying ID lookup first (PRIMARY KEY),
        then name lookup (idx_vm_name index) if not found.

        Args:
            identifier: VM ID or name

        Returns:
            True if removed, False if VM not found
        """
        with self._get_pooled_connection(readonly=False) as conn:
            # Try ID first (PRIMARY KEY index - O(log n))
            cursor = conn.execute(
                "DELETE FROM vm_instances WHERE id = ?",
                (identifier,),
            )

            if cursor.rowcount > 0:
                return True

            # Try name (idx_vm_name index - O(log n))
            cursor = conn.execute(
                "DELETE FROM vm_instances WHERE name = ?",
                (identifier,),
            )

            return cursor.rowcount > 0

    def update_vm_config(
        self, identifier: str, new_config: Dict[str, Any]
    ) -> bool:
        """
        Update VM configuration in database.

        Optimized to use indexes by trying ID lookup first (PRIMARY KEY),
        then name lookup (idx_vm_name index) if not found.

        Args:
            identifier: VM name or ID
            new_config: New configuration data

        Returns:
            True if update successful, False if VM not found

        Raises:
            StateError: If database operation fails
        """
        try:
            with self._get_pooled_connection(readonly=False) as conn:
                # Try ID first (PRIMARY KEY index - O(log n))
                cursor = conn.execute(
                    """
                    UPDATE vm_instances
                    SET config_data = ?
                    WHERE id = ?
                """,
                    (json.dumps(new_config), identifier),
                )

                if cursor.rowcount > 0:
                    return True

                # Try name (idx_vm_name index - O(log n))
                cursor = conn.execute(
                    """
                    UPDATE vm_instances
                    SET config_data = ?
                    WHERE name = ?
                """,
                    (json.dumps(new_config), identifier),
                )

                return cursor.rowcount > 0

        except sqlite3.Error as e:
            raise StateError(f"Database error updating VM config: {e}")
        except Exception as e:
            raise StateError(f"Error updating VM config: {e}")

    def _row_to_vm_instance(self, row: sqlite3.Row) -> "VMInstance":
        """Convert database row to VMInstance object."""
        # Import at runtime to avoid circular dependency
        from ..state import VMInstance

        config_data = json.loads(row["config_data"])

        # Handle runner_pid which may not exist in older schemas
        runner_pid = row["runner_pid"] if "runner_pid" in row.keys() else None

        # Handle auth_secret which may not exist in older schemas
        auth_secret = row["auth_secret"] if "auth_secret" in row.keys() else None

        # Handle qmp_socket_path which may not exist in older schemas
        qmp_socket_path = row["qmp_socket_path"] if "qmp_socket_path" in row.keys() else None

        return VMInstance(
            id=row["id"],
            name=row["name"],
            config_path=row["config_path"],
            config_data=config_data,
            status=row["status"],
            pid=row["pid"],
            runner_pid=runner_pid,
            socket_path=row["socket_path"],
            qmp_socket_path=qmp_socket_path,
            created_at=datetime.fromisoformat(row["created_at"]),
            updated_at=datetime.fromisoformat(row["updated_at"]),
            auth_secret=auth_secret,
        )
